import configparser
import json

from typing import Any, Dict, Optional, Tuple, Union
from urllib.parse import urlencode

from authlib.oidc.core import CodeIDToken
from authlib.jose import jwt
from authlib.jose.errors import ExpiredTokenError

from safety.auth.models import Organization
from safety.auth.constants import AUTH_SERVER_URL, CLI_AUTH_LOGOUT, CLI_CALLBACK, AUTH_CONFIG_USER, CLI_AUTH
from safety.constants import CONFIG
from safety.errors import NotVerifiedEmailError
from safety.scan.util import Stage
from safety.util import get_proxy_dict


def get_authorization_data(client, code_verifier: str,
                           organization: Optional[Organization] = None, 
                           sign_up: bool = False, ensure_auth: bool = False, headless: bool = False) -> Tuple[str, str]:
    
    kwargs = {'sign_up': sign_up, 'locale': 'en', 'ensure_auth': ensure_auth, 'headless': headless}
    if organization:
        kwargs['organization'] = organization.id

    return client.create_authorization_url(CLI_AUTH,
                                           code_verifier=code_verifier,
                                           **kwargs)

def get_logout_url(id_token: str) -> str:
    return f'{CLI_AUTH_LOGOUT}?id_token={id_token}'

def get_redirect_url() -> str:
    return CLI_CALLBACK

def get_organization() -> Optional[Organization]:
    config = configparser.ConfigParser()
    config.read(CONFIG)

    org_conf: Union[Dict[str, str], configparser.SectionProxy] = config[
        'organization'] if 'organization' in config.sections() else {}
    org_id: Optional[str] = org_conf['id'].replace("\"", "") if org_conf.get('id', None) else None
    org_name: Optional[str] = org_conf['name'].replace("\"", "") if org_conf.get('name', None) else None

    if not org_id:
        return None

    org = Organization(
        id=org_id,
        name=org_name
    )

    return org

def get_auth_info(ctx):
    from safety.auth.utils import is_email_verified
    
    info = None
    if ctx.obj.auth.client.token:
        try:
            info = get_token_data(get_token(name='id_token'), keys=ctx.obj.auth.keys)

            verified = is_email_verified(info)
            if not verified:
                user_info = ctx.obj.auth.client.fetch_user_info()
                verified = is_email_verified(user_info)

                if verified:
                    # refresh only if needed 
                    raise ExpiredTokenError

        except ExpiredTokenError as e:
            # id_token expired. So fire a manually a refresh
            try:
                ctx.obj.auth.client.refresh_token(ctx.obj.auth.client.metadata.get('token_endpoint'),
                                         refresh_token=ctx.obj.auth.client.token.get('refresh_token'))
                info = get_token_data(get_token(name='id_token'), keys=ctx.obj.auth.keys)
            except Exception as _e:
                clean_session(ctx.obj.auth.client)
        except Exception as _g:
            clean_session(ctx.obj.auth.client)
    
    return info

def get_token_data(token, keys, silent_if_expired=False) -> Optional[Dict]:
    claims = jwt.decode(token, keys, claims_cls=CodeIDToken)
    try:
        claims.validate()
    except ExpiredTokenError as e:
        if not silent_if_expired:
            raise e

    return claims

def get_token(name='access_token') -> Optional[str]:
    """"
    This returns tokens saved in the local auth configuration.
    There are two types of tokens: access_token and id_token
    """
    config = configparser.ConfigParser()
    config.read(AUTH_CONFIG_USER)

    if 'auth' in config.sections() and name in config['auth']:
        value = config['auth'][name]
        if value:
            return value

    return None

def get_host_config(key_name) -> Optional[Any]:
    config = configparser.ConfigParser()
    config.read(CONFIG)

    if not config.has_section("host"):
        return None
    
    host_section = dict(config.items("host"))

    if key_name in host_section:
        if key_name == "stage":
            # Support old alias in the config.ini
            if host_section[key_name] == "dev":
                host_section[key_name] = "development"
            if host_section[key_name] not in {env.value for env in Stage}:
                return None
            return Stage(host_section[key_name])

    return None

def str_to_bool(s):
    """Convert a string to a boolean value."""
    if s.lower() == 'true' or s == '1':
        return True
    elif s.lower() == 'false' or s == '0':
        return False
    else:
        raise ValueError(f"Cannot convert '{s}' to a boolean value.")

def get_proxy_config() -> Tuple[Dict[str, str], Optional[int], bool]:
    config = configparser.ConfigParser()
    config.read(CONFIG)

    proxy_dictionary =  None
    required = False
    timeout = None
    proxy = None

    if config.has_section("proxy"):
        proxy = dict(config.items("proxy"))

    if proxy:
        try:
            proxy_dictionary = get_proxy_dict(proxy['protocol'], proxy['host'], 
                                                proxy['port'])
            required = str_to_bool(proxy["required"])
            timeout = proxy["timeout"]
        except Exception as e:
            pass

    return proxy_dictionary, timeout, required

def clean_session(client):
    config = configparser.ConfigParser()
    config['auth'] = {'access_token': '', 'id_token': '', 'refresh_token':''}

    with open(AUTH_CONFIG_USER, 'w') as configfile:
        config.write(configfile)

    client.token = None

    return True

def save_auth_config(access_token=None, id_token=None, refresh_token=None):
    config = configparser.ConfigParser()
    config.read(AUTH_CONFIG_USER)
    config['auth'] = {'access_token': access_token, 'id_token': id_token, 
                      'refresh_token': refresh_token}
    
    with open(AUTH_CONFIG_USER, 'w') as configfile:
        config.write(configfile)




import logging
import os
from pathlib import Path
import re
from typing import Dict, List, Optional, Set, Tuple, Union
from safety_schemas.models import Ecosystem, FileType

from safety.errors import SafetyException

from .handlers import FileHandler, ECOSYSTEM_HANDLER_MAPPING

LOG = logging.getLogger(__name__)

def should_exclude(excludes: Set[Path], to_analyze: Path) -> bool:

    if not to_analyze.is_absolute():
        to_analyze = to_analyze.resolve()

    for exclude in excludes:
        if not exclude.is_absolute():
            exclude = exclude.resolve()

        try:
            if to_analyze == exclude or \
                to_analyze.relative_to(exclude):
                return True
        except ValueError:
            pass
    
    return False


class FileFinder():
    """"
    Defines a common interface to agree in what type of components Safety is trying to
    find depending on the language type.
    """

    def __init__(self, max_level: int, ecosystems: List[Ecosystem], target: Path, 
                 console, live_status=None,
                 exclude: Optional[List[str]] = None, 
                 include_files: Optional[Dict[FileType, List[Path]]] = None,
                 handlers: Optional[Set[FileHandler]] = None) -> None:
        self.max_level = max_level
        self.target = target
        self.include_files = include_files

        if not handlers:
            handlers = set(ECOSYSTEM_HANDLER_MAPPING[ecosystem]() 
                                              for ecosystem in ecosystems)
        
        self.handlers = handlers
        self.file_count = 0
        self.exclude_dirs: Set[Path] = set()
        self.exclude_files: Set[Path] = set()
        exclude = [] if not exclude else exclude

        for pattern in exclude:
            for path in Path(target).glob(pattern):
                if path.is_dir():
                    self.exclude_dirs.add(path)
                else:
                    self.exclude_files.add(path)

        self.console = console
        self.live_status = live_status
    
    def process_directory(self, dir_path, max_deep: Optional[int]=None) -> Tuple[str, Dict[str, Set[Path]]]:
        files: Dict[str, Set[Path]] = {}
        level : int = 0
        initial_depth = len(Path(dir_path).parts) - 1
        skip_dirs = set()
        skip_files = set()

        for root, dirs, filenames in os.walk(dir_path):
            root_path = Path(root)
            current_depth = len(root_path.parts) - initial_depth

            dirs[:] = [d for d in dirs if not should_exclude(excludes=self.exclude_dirs,
                                                             to_analyze=(root_path / Path(d)))]
            
            if dirs:
                LOG.info(f"Directories to inspect -> {', '.join(dirs)}")
            
            LOG.info(f"Current -> {root}")
            if self.live_status:
                self.live_status.update(f":mag: Scanning {root}")

            if max_deep is not None and current_depth > max_deep:
                # Don't go deeper
                del dirs[:]

            filenames[:] = [f for f in filenames if not should_exclude(
                excludes=self.exclude_files, 
                to_analyze=Path(f))]

            self.file_count += len(filenames)

            for file_name in filenames:
                for handler in self.handlers:
                    file_type = handler.can_handle(root, file_name, self.include_files)
                    if file_type:
                        inspectable_file: Path = Path(root, file_name)
                        if file_type.value not in files or not files[file_type.value]:
                            files[file_type.value] = set()
                        files[file_type.value].add(inspectable_file)
                        break
            level += 1

        return dir_path, files

    def search(self) -> Tuple[str, Dict[str, Set[Path]]]:
        return self.process_directory(self.target, self.max_level)



from collections import defaultdict
import logging
import sys
from typing import Any, DefaultDict, Dict, List, Optional, Tuple, Union
import click
from functools import wraps
import typer
from typer.core import TyperGroup, TyperCommand, MarkupMode
from safety.auth.constants import CLI_AUTH, MSG_NON_AUTHENTICATED
from safety.auth.models import Auth
from safety.constants import MSG_NO_AUTHD_CICD_PROD_STG, MSG_NO_AUTHD_CICD_PROD_STG_ORG, MSG_NO_AUTHD_DEV_STG, MSG_NO_AUTHD_DEV_STG_ORG_PROMPT, MSG_NO_AUTHD_DEV_STG_PROMPT, MSG_NO_AUTHD_NOTE_CICD_PROD_STG_TPL, MSG_NO_VERIFIED_EMAIL_TPL
from safety.scan.constants import CONSOLE_HELP_THEME

from safety.scan.models import ScanOutput

from .util import output_exception
from .errors import SafetyError, SafetyException

LOG = logging.getLogger(__name__)


def get_command_for(name:str, typer_instance: typer.Typer):
    single_command = next(
        (command 
         for command in typer_instance.registered_commands 
         if command.name == name), None)

    if not single_command:
        raise ValueError("Unable to find the command name.")
    
    single_command.context_settings = typer_instance.info.context_settings
    click_command = typer.main.get_command_from_info(
        single_command,
        pretty_exceptions_short=typer_instance.pretty_exceptions_short,
        rich_markup_mode=typer_instance.rich_markup_mode,
    )
    if typer_instance._add_completion:
        click_install_param, click_show_param = \
            typer.main.get_install_completion_arguments()
        click_command.params.append(click_install_param)
        click_command.params.append(click_show_param)
    return click_command


def pass_safety_cli_obj(func):
    """
    Make sure the SafetyCLI object exists for a command.
    """
    @wraps(func)
    def inner(ctx, *args, **kwargs):

        if not ctx.obj:
            from .models import SafetyCLI
            ctx.obj = SafetyCLI()

        return func(ctx, *args, **kwargs)

    return inner


def pretty_format_help(obj: Union[click.Command, click.Group], 
                       ctx: click.Context, markup_mode: MarkupMode) -> None:
    from typer.rich_utils import _print_options_panel, _get_rich_console, \
        _get_help_text, highlighter, STYLE_HELPTEXT, STYLE_USAGE_COMMAND, _print_commands_panel, \
            _RICH_HELP_PANEL_NAME, ARGUMENTS_PANEL_TITLE, OPTIONS_PANEL_TITLE, \
                COMMANDS_PANEL_TITLE, _make_rich_rext
    from rich.align import Align
    from rich.padding import Padding
    from rich.console import Console
    from rich.theme import Theme

    typer_console = _get_rich_console()

    with typer_console.use_theme(Theme(styles=CONSOLE_HELP_THEME)) as theme_context:
        console = theme_context.console
        # Print command / group help if we have some
        if obj.help:
            console.print()

            # Print with some padding            
            console.print(
                Padding(
                    Align(_get_help_text(obj=obj, markup_mode=markup_mode), pad=False),
                    (0, 1, 0, 1)
                )
            )

        # Print usage
        console.print(
            Padding(highlighter(obj.get_usage(ctx)), 1), style=STYLE_USAGE_COMMAND
        )

        if isinstance(obj, click.MultiCommand):
            panel_to_commands: DefaultDict[str, List[click.Command]] = defaultdict(list)
            for command_name in obj.list_commands(ctx):
                command = obj.get_command(ctx, command_name)
                if command and not command.hidden:
                    panel_name = (
                        getattr(command, _RICH_HELP_PANEL_NAME, None)
                        or COMMANDS_PANEL_TITLE
                    )
                    panel_to_commands[panel_name].append(command)

            # Print each command group panel
            default_commands = panel_to_commands.get(COMMANDS_PANEL_TITLE, [])
            _print_commands_panel(
                name=COMMANDS_PANEL_TITLE,
                commands=default_commands,
                markup_mode=markup_mode,
                console=console,
            )
            for panel_name, commands in panel_to_commands.items():
                if panel_name == COMMANDS_PANEL_TITLE:
                    # Already printed above
                    continue
                _print_commands_panel(
                    name=panel_name,
                    commands=commands,
                    markup_mode=markup_mode,
                    console=console,
                )        

        panel_to_arguments: DefaultDict[str, List[click.Argument]] = defaultdict(list)
        panel_to_options: DefaultDict[str, List[click.Option]] = defaultdict(list)
        for param in obj.get_params(ctx):
            # Skip if option is hidden
            if getattr(param, "hidden", False):
                continue
            if isinstance(param, click.Argument):
                panel_name = (
                    getattr(param, _RICH_HELP_PANEL_NAME, None) or ARGUMENTS_PANEL_TITLE
                )
                panel_to_arguments[panel_name].append(param)
            elif isinstance(param, click.Option):
                panel_name = (
                    getattr(param, _RICH_HELP_PANEL_NAME, None) or OPTIONS_PANEL_TITLE
                )
                panel_to_options[panel_name].append(param)

        default_options = panel_to_options.get(OPTIONS_PANEL_TITLE, [])
        _print_options_panel(
            name=OPTIONS_PANEL_TITLE,
            params=default_options,
            ctx=ctx,
            markup_mode=markup_mode,
            console=console,
        )
        for panel_name, options in panel_to_options.items():
            if panel_name == OPTIONS_PANEL_TITLE:
                # Already printed above
                continue
            _print_options_panel(
                name=panel_name,
                params=options,
                ctx=ctx,
                markup_mode=markup_mode,
                console=console,
            )

        default_arguments = panel_to_arguments.get(ARGUMENTS_PANEL_TITLE, [])
        _print_options_panel(
            name=ARGUMENTS_PANEL_TITLE,
            params=default_arguments,
            ctx=ctx,
            markup_mode=markup_mode,
            console=console,
        )
        for panel_name, arguments in panel_to_arguments.items():
            if panel_name == ARGUMENTS_PANEL_TITLE:
                # Already printed above
                continue
            _print_options_panel(
                name=panel_name,
                params=arguments,
                ctx=ctx,
                markup_mode=markup_mode,
                console=console,
            )

        if ctx.parent:
            params = []
            for param in ctx.parent.command.params:
                if isinstance(param, click.Option):
                    params.append(param)

            _print_options_panel(
                name="Global-Options",
                params=params,
                ctx=ctx.parent,
                markup_mode=markup_mode,
                console=console,
            )            

        # Epilogue if we have it
        if obj.epilog:
            # Remove single linebreaks, replace double with single
            lines = obj.epilog.split("\n\n")
            epilogue = "\n".join([x.replace("\n", " ").strip() for x in lines])
            epilogue_text = _make_rich_rext(text=epilogue, markup_mode=markup_mode)
            console.print(Padding(Align(epilogue_text, pad=False), 1))



def print_main_command_panels(*,
    name: str,
    commands: List[click.Command],
    markup_mode: MarkupMode,
    console):
    from rich import box
    from rich.table import Table
    from rich.text import Text
    from rich.panel import Panel
    from typer.rich_utils import STYLE_COMMANDS_TABLE_SHOW_LINES, STYLE_COMMANDS_TABLE_LEADING, \
        STYLE_COMMANDS_TABLE_BOX, STYLE_COMMANDS_TABLE_BORDER_STYLE, STYLE_COMMANDS_TABLE_ROW_STYLES, \
            STYLE_COMMANDS_TABLE_PAD_EDGE, STYLE_COMMANDS_TABLE_PADDING, STYLE_COMMANDS_PANEL_BORDER, \
                ALIGN_COMMANDS_PANEL, _make_command_help

    t_styles: Dict[str, Any] = {
        "show_lines": STYLE_COMMANDS_TABLE_SHOW_LINES,
        "leading": STYLE_COMMANDS_TABLE_LEADING,
        "box": STYLE_COMMANDS_TABLE_BOX,
        "border_style": STYLE_COMMANDS_TABLE_BORDER_STYLE,
        "row_styles": STYLE_COMMANDS_TABLE_ROW_STYLES,
        "pad_edge": STYLE_COMMANDS_TABLE_PAD_EDGE,
        "padding": STYLE_COMMANDS_TABLE_PADDING,
    }
    box_style = getattr(box, t_styles.pop("box"), None)

    commands_table = Table(
        highlight=False,
        show_header=False,
        expand=True,
        box=box_style,
        **t_styles,
    )

    console_width = 80
    column_width = 25

    if console.size and console.size[0] > 80:
        console_width = console.size[0]

    commands_table.add_column(style="bold cyan", no_wrap=True, width=column_width, max_width=column_width)
    commands_table.add_column(width=console_width - column_width)
    
    rows = []

    for command in commands:
        helptext = command.short_help or command.help or ""
        command_name = command.name or ""
        command_name_text = Text(command_name)
        rows.append(
            [
                command_name_text,
                _make_command_help(
                    help_text=helptext,
                    markup_mode=markup_mode,
                ),
            ]
        )
        rows.append([])
    for row in rows:
        commands_table.add_row(*row)
    if commands_table.row_count:
        console.print(
            Panel(
                commands_table,
                border_style=STYLE_COMMANDS_PANEL_BORDER,
                title=name,
                title_align=ALIGN_COMMANDS_PANEL,
            )
        )

# The help output for the main safety root command: `safety --help`
def format_main_help(obj: Union[click.Command, click.Group],
                       ctx: click.Context, markup_mode: MarkupMode) -> None:
    from typer.rich_utils import _print_options_panel, _get_rich_console, \
    _get_help_text, highlighter, STYLE_USAGE_COMMAND, _print_commands_panel, \
        _RICH_HELP_PANEL_NAME, ARGUMENTS_PANEL_TITLE, OPTIONS_PANEL_TITLE, \
            COMMANDS_PANEL_TITLE, _make_rich_rext
    from rich.align import Align
    from rich.padding import Padding
    from rich.console import Console
    from rich.theme import Theme

    typer_console = _get_rich_console()

    with typer_console.use_theme(Theme(styles=CONSOLE_HELP_THEME)) as theme_context:
        console = theme_context.console

        # Print command / group help if we have some
        if obj.help:
            console.print()
            # Print with some padding
            console.print(
                Padding(
                    Align(_get_help_text(obj=obj, markup_mode=markup_mode, ),
                        pad=False,
                    ),
                    (0, 1, 0, 1),
                )
            )

        # Print usage
        console.print(
            Padding(highlighter(obj.get_usage(ctx)), 1), style=STYLE_USAGE_COMMAND
        )

        if isinstance(obj, click.MultiCommand):
            UTILITY_COMMANDS_PANEL_TITLE = "Commands cont."

            panel_to_commands: DefaultDict[str, List[click.Command]] = defaultdict(list)
            for command_name in obj.list_commands(ctx):
                command = obj.get_command(ctx, command_name)
                if command and not command.hidden:
                    panel_name = (
                        UTILITY_COMMANDS_PANEL_TITLE if command.utility_command else COMMANDS_PANEL_TITLE
                    )
                    panel_to_commands[panel_name].append(command)

            # Print each command group panel
            default_commands = panel_to_commands.get(COMMANDS_PANEL_TITLE, [])
            print_main_command_panels(
                name=COMMANDS_PANEL_TITLE,
                commands=default_commands,
                markup_mode=markup_mode,
                console=console,
            )
            for panel_name, commands in panel_to_commands.items():
                if panel_name == COMMANDS_PANEL_TITLE:
                    # Already printed above
                    continue
                print_main_command_panels(
                    name=panel_name,
                    commands=commands,
                    markup_mode=markup_mode,
                    console=console,
                )

        panel_to_arguments: DefaultDict[str, List[click.Argument]] = defaultdict(list)
        panel_to_options: DefaultDict[str, List[click.Option]] = defaultdict(list)
        for param in obj.get_params(ctx):
            # Skip if option is hidden
            if getattr(param, "hidden", False):
                continue
            if isinstance(param, click.Argument):
                panel_name = (
                    getattr(param, _RICH_HELP_PANEL_NAME, None) or ARGUMENTS_PANEL_TITLE
                )
                panel_to_arguments[panel_name].append(param)
            elif isinstance(param, click.Option):
                panel_name = (
                    getattr(param, _RICH_HELP_PANEL_NAME, None) or OPTIONS_PANEL_TITLE
                )
                panel_to_options[panel_name].append(param)
        default_arguments = panel_to_arguments.get(ARGUMENTS_PANEL_TITLE, [])
        _print_options_panel(
            name=ARGUMENTS_PANEL_TITLE,
            params=default_arguments,
            ctx=ctx,
            markup_mode=markup_mode,
            console=console,
        )
        for panel_name, arguments in panel_to_arguments.items():
            if panel_name == ARGUMENTS_PANEL_TITLE:
                # Already printed above
                continue
            _print_options_panel(
                name=panel_name,
                params=arguments,
                ctx=ctx,
                markup_mode=markup_mode,
                console=console,
            )
        default_options = panel_to_options.get(OPTIONS_PANEL_TITLE, [])
        _print_options_panel(
            name=OPTIONS_PANEL_TITLE,
            params=default_options,
            ctx=ctx,
            markup_mode=markup_mode,
            console=console,
        )
        for panel_name, options in panel_to_options.items():
            if panel_name == OPTIONS_PANEL_TITLE:
                # Already printed above
                continue
            _print_options_panel(
                name=panel_name,
                params=options,
                ctx=ctx,
                markup_mode=markup_mode,
                console=console,
            )

        # Epilogue if we have it
        if obj.epilog:
            # Remove single linebreaks, replace double with single
            lines = obj.epilog.split("\n\n")
            epilogue = "\n".join([x.replace("\n", " ").strip() for x in lines])
            epilogue_text = _make_rich_rext(text=epilogue, markup_mode=markup_mode)
            console.print(Padding(Align(epilogue_text, pad=False), 1))


def process_auth_status_not_ready(console, auth: Auth, ctx: typer.Context):
    from safety_schemas.models import Stage
    from rich.prompt import Confirm, Prompt

    if not auth.client or not auth.client.is_using_auth_credentials():

        if auth.stage is Stage.development:
            console.print()
            if auth.org:
                confirmed = Confirm.ask(MSG_NO_AUTHD_DEV_STG_ORG_PROMPT, choices=["Y", "N", "y", "n"], 
                                        show_choices=False, show_default=False, 
                                        default=True, console=console)
                
                if not confirmed:
                    sys.exit(0)

                from safety.auth.cli import auth_app
                login_command = get_command_for(name='login',
                                        typer_instance=auth_app)
                ctx.invoke(login_command)
            else:
                console.print(MSG_NO_AUTHD_DEV_STG)
                console.print()
                choices = ["L", "R", "l", "r"]
                next_command = Prompt.ask(MSG_NO_AUTHD_DEV_STG_PROMPT, default=None, 
                                          choices=choices, show_choices=False, 
                                          console=console)
                
                from safety.auth.cli import auth_app
                login_command = get_command_for(name='login',
                                        typer_instance=auth_app)
                register_command = get_command_for(name='register',
                                        typer_instance=auth_app)
                if next_command is None or next_command.lower() not in choices:
                    sys.exit(0)
                
                console.print()                
                if next_command.lower() == "r":
                    ctx.invoke(register_command)
                else:
                    ctx.invoke(login_command)

                if not ctx.obj.auth.email_verified:
                    sys.exit(1)
        else:
            if not auth.org:
                console.print(MSG_NO_AUTHD_CICD_PROD_STG_ORG.format(LOGIN_URL=CLI_AUTH))
            
            else:
                console.print(MSG_NO_AUTHD_CICD_PROD_STG)
                console.print(
                    MSG_NO_AUTHD_NOTE_CICD_PROD_STG_TPL.format(
                        LOGIN_URL=CLI_AUTH, 
                        SIGNUP_URL=f"{CLI_AUTH}/?sign_up=True"))
            sys.exit(1)

    elif not auth.email_verified:
        console.print()
        console.print(MSG_NO_VERIFIED_EMAIL_TPL.format(email=auth.email if auth.email else "Missing email"))
        sys.exit(1)
    else:
        console.print(MSG_NON_AUTHENTICATED)
        sys.exit(1)

class UtilityCommandMixin:
    def __init__(self, *args, **kwargs):
        self.utility_command = kwargs.pop('utility_command', False)
        super().__init__(*args, **kwargs)

class SafetyCLISubGroup(UtilityCommandMixin, TyperGroup):

    def format_help(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        pretty_format_help(self, ctx, markup_mode=self.rich_markup_mode)

    def format_usage(self, ctx, formatter) -> None:
        command_path = ctx.command_path
        pieces = self.collect_usage_pieces(ctx)
        main_group = ctx.parent
        if main_group:
            command_path = f"{main_group.command_path} [GLOBAL-OPTIONS] {ctx.command.name}"

        formatter.write_usage(command_path, " ".join(pieces))

    def command(
        self,
        *args: Any,
        **kwargs: Any,
    ):
        super().command(*args, **kwargs)

class SafetyCLICommand(UtilityCommandMixin, TyperCommand):

    def format_help(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        pretty_format_help(self, ctx, markup_mode=self.rich_markup_mode)

    def format_usage(self, ctx, formatter) -> None:
        command_path = ctx.command_path
        pieces = self.collect_usage_pieces(ctx)
        main_group = ctx.parent
        if main_group:
            command_path = f"{main_group.command_path} [GLOBAL-OPTIONS] {ctx.command.name}"

        formatter.write_usage(command_path, " ".join(pieces))


class SafetyCLIUtilityCommand(TyperCommand):
    def __init__(self, *args, **kwargs):
        self.utility_command = True
        super().__init__(*args, **kwargs)

class SafetyCLILegacyGroup(UtilityCommandMixin, click.Group):

    def parse_legacy_args(self, args: List[str]) -> Tuple[Optional[Dict[str, str]], Optional[str]]:
        options = {
            'proxy_protocol': 'https',
            'proxy_port': 80,
            'proxy_host': None
        }
        key = None

        for i, arg in enumerate(args):
            if arg in ['--proxy-protocol', '-pr'] and i + 1 < len(args):
                options['proxy_protocol'] = args[i + 1]
            elif arg in ['--proxy-port', '-pp'] and i + 1 < len(args):
                options['proxy_port'] = int(args[i + 1])
            elif arg in ['--proxy-host', '-ph'] and i + 1 < len(args):
                options['proxy_host'] = args[i + 1]
            elif arg in ['--key'] and i + 1 < len(args):
                key = args[i + 1]

        proxy = options if options['proxy_host'] else None
        return proxy, key

    def invoke(self, ctx):
        args = ctx.args

        # Workaround for legacy check options, that now are global options
        subcommand_args = set(args)
        PROXY_HOST_OPTIONS = set(["--proxy-host", "-ph"])
        if "check" in ctx.protected_args or "license" in ctx.protected_args and (bool(PROXY_HOST_OPTIONS.intersection(subcommand_args) or "--key" in subcommand_args)) :
            proxy_options, key = self.parse_legacy_args(args)
            if proxy_options:
                ctx.params.update(proxy_options)
            
            if key:
                ctx.params.update({"key": key})

        # Now, invoke the original behavior
        super(SafetyCLILegacyGroup, self).invoke(ctx)
    

    def format_help(self, ctx, formatter) -> None:
        # The main `safety --help`
        if self.name == "cli":
            format_main_help(self, ctx, markup_mode="rich")
        # All other help outputs
        else: 
            pretty_format_help(self, ctx, markup_mode="rich")

class SafetyCLILegacyCommand(UtilityCommandMixin, click.Command):
    def format_help(self, ctx, formatter) -> None:
        pretty_format_help(self, ctx, markup_mode="rich")


def handle_cmd_exception(func):
    @wraps(func)
    def inner(ctx, output: Optional[ScanOutput], *args, **kwargs):
        if output:
            kwargs.update({"output": output})
        
            if output is ScanOutput.NONE:
                return func(ctx, *args, **kwargs)

        try:
            return func(ctx, *args, **kwargs)
        except click.ClickException as e:
            raise e
        except SafetyError as e:
            LOG.exception('Expected SafetyError happened: %s', e)
            output_exception(e, exit_code_output=True)
        except Exception as e:
            LOG.exception('Unexpected Exception happened: %s', e)
            exception = e if isinstance(e, SafetyException) else SafetyException(info=e)
            output_exception(exception, exit_code_output=True)        

    return inner


from typing import Dict, Generic, TypeVar

from pydantic import BaseModel as PydanticBaseModel
from pydantic import Extra
from pydantic.validators import dict_validator

from common.const import SCHEMA_DICT_ITEMS_COUNT_LIMIT
from common.exceptions import DictMaxLengthError


class BaseModel(PydanticBaseModel):
    class Config:
        arbitrary_types_allowed = True
        max_anystr_length = 50
        validate_assignment = True
        extra = Extra.forbid


KeyType = TypeVar("KeyType")
ValueType = TypeVar("ValueType")


class ConstrainedDict(Generic[KeyType, ValueType]):
    def __init__(self, v: Dict[KeyType, ValueType]):
        super().__init__()

    @classmethod
    def __get_validators__(cls):
        yield cls.dict_length_validator

    @classmethod
    def dict_length_validator(cls, v):
        v = dict_validator(v)
        if len(v) > SCHEMA_DICT_ITEMS_COUNT_LIMIT:
            raise DictMaxLengthError(limit_value=SCHEMA_DICT_ITEMS_COUNT_LIMIT)
        return v



# -*- coding: utf-8 -*-
from dataclasses import asdict
import errno
import itertools
import json
import logging
import os
from pathlib import Path
import random
import sys
import tempfile
import time
from collections import defaultdict
from datetime import datetime
from typing import Dict, Optional, List, Any

import click
import requests
from requests.models import PreparedRequest
from packaging.specifiers import SpecifierSet
from packaging.utils import canonicalize_name
from packaging.version import parse as parse_version, Version
from pydantic.json import pydantic_encoder
from filelock import FileLock

from safety_schemas.models import Ecosystem, FileType



from .constants import (API_MIRRORS, DB_CACHE_FILE, OPEN_MIRRORS, REQUEST_TIMEOUT, DATA_API_BASE_URL, JSON_SCHEMA_VERSION,
                        IGNORE_UNPINNED_REQ_REASON)
from .errors import (DatabaseFetchError, DatabaseFileNotFoundError, InvalidCredentialError,
                     TooManyRequestsError, NetworkConnectionError,
                     RequestTimeoutError, ServerError, MalformedDatabase)
from .models import Vulnerability, CVE, Severity, Fix, is_pinned_requirement, SafetyRequirement
from .output_utils import print_service, get_applied_msg, prompt_service, get_skipped_msg, get_fix_opt_used_msg, \
    is_using_api_key, get_specifier_range_info
from .util import build_remediation_info_url, pluralize, read_requirements, Package, build_telemetry_data, sync_safety_context, \
    SafetyContext, validate_expiration_date, is_a_remote_mirror, get_requirements_content, SafetyPolicyFile, \
    get_terminal_size, is_ignore_unpinned_mode, get_hashes

LOG = logging.getLogger(__name__)


def get_from_cache(db_name: str, cache_valid_seconds: int = 0, skip_time_verification: bool = False) -> Optional[Dict[str, Any]]:
    cache_file_lock = f"{DB_CACHE_FILE}.lock"
    os.makedirs(os.path.dirname(cache_file_lock), exist_ok=True)
    lock = FileLock(cache_file_lock, timeout=10)
    with lock:
        if os.path.exists(DB_CACHE_FILE):
            with open(DB_CACHE_FILE) as f:
                try:
                    data = json.loads(f.read())
                    if db_name in data:

                        if "cached_at" in data[db_name]:
                            if data[db_name]["cached_at"] + cache_valid_seconds > time.time() or skip_time_verification:
                                LOG.debug('Getting the database from cache at %s, cache setting: %s',
                                        data[db_name]["cached_at"], cache_valid_seconds)

                                try:
                                    data[db_name]["db"]["meta"]["base_domain"] = "https://data.safetycli.com"
                                except KeyError as e:
                                    pass

                                return data[db_name]["db"]

                            LOG.debug('Cached file is too old, it was cached at %s', data[db_name]["cached_at"])
                        else:
                            LOG.debug('There is not the cached_at key in %s database', data[db_name])

                except json.JSONDecodeError:
                    LOG.debug('JSONDecodeError trying to get the cached database.')
        else:
            LOG.debug("Cache file doesn't exist...")
    return None


def write_to_cache(db_name, data):
    # cache is in: ~/safety/cache.json
    # and has the following form:
    # {
    #   "insecure.json": {
    #       "cached_at": 12345678
    #       "db": {}
    #   },
    #   "insecure_full.json": {
    #       "cached_at": 12345678
    #       "db": {}
    #   },
    # }
    if not os.path.exists(os.path.dirname(DB_CACHE_FILE)):
        try:
            os.makedirs(os.path.dirname(DB_CACHE_FILE))
            with open(DB_CACHE_FILE, "w") as _:
                _.write(json.dumps({}))
                LOG.debug('Cache file created')
        except OSError as exc:  # Guard against race condition
            LOG.debug('Unable to create the cache file because: %s', exc.errno)
            if exc.errno != errno.EEXIST:
                raise

    cache_file_lock = f"{DB_CACHE_FILE}.lock"
    lock = FileLock(cache_file_lock, timeout=10)
    with lock:
        if os.path.exists(DB_CACHE_FILE):
            with open(DB_CACHE_FILE, "r") as f:
                try:
                    cache = json.loads(f.read())
                except json.JSONDecodeError:
                    LOG.debug('JSONDecodeError in the local cache, dumping the full cache file.')
                    cache = {}
        else:
            cache = {}

        with open(DB_CACHE_FILE, "w") as f:
            cache[db_name] = {
                "cached_at": time.time(),
                "db": data
            }
            f.write(json.dumps(cache))
            LOG.debug('Safety updated the cache file for %s database.', db_name)


def fetch_database_url(session, mirror, db_name, cached, telemetry=True,
                       ecosystem: Ecosystem = Ecosystem.PYTHON, from_cache=True):
    headers = {'schema-version': JSON_SCHEMA_VERSION, 'ecosystem': ecosystem.value}

    if cached and from_cache:
        cached_data = get_from_cache(db_name=db_name, cache_valid_seconds=cached)
        if cached_data:
            LOG.info('Database %s returned from cache.', db_name)
            return cached_data
    url = mirror + db_name


    telemetry_data = {
        'telemetry': json.dumps(build_telemetry_data(telemetry=telemetry),
                                default=pydantic_encoder)}

    try:
        r = session.get(url=url, timeout=REQUEST_TIMEOUT,
                        headers=headers, params=telemetry_data)
    except requests.exceptions.ConnectionError:
        raise NetworkConnectionError()
    except requests.exceptions.Timeout:
        raise RequestTimeoutError()
    except requests.exceptions.RequestException:
        raise DatabaseFetchError()

    if r.status_code == 403:
        raise InvalidCredentialError(credential=session.get_credential(), reason=r.text)

    if r.status_code == 429:
        raise TooManyRequestsError(reason=r.text)

    if r.status_code != 200:
        raise ServerError(reason=r.reason)

    try:
        data = r.json()
    except json.JSONDecodeError as e:
        raise MalformedDatabase(reason=e)

    if cached:
        LOG.info('Writing %s to cache because cached value was %s', db_name, cached)
        write_to_cache(db_name, data)

    return data


def fetch_policy(session):
    url = f"{DATA_API_BASE_URL}policy/"

    try:
        LOG.debug(f'Getting policy')
        r = session.get(url=url, timeout=REQUEST_TIMEOUT)
        LOG.debug(r.text)
        return r.json()
    except Exception:
        LOG.exception("Error fetching policy")

        return {"safety_policy": "", "audit_and_monitor": False}


def post_results(session, safety_json, policy_file):
    url = f"{DATA_API_BASE_URL}result/"

    # safety_json is in text form already. policy_file is a text YAML
    audit_report = {
        "safety_json": json.loads(safety_json),
        "policy_file": policy_file
    }

    try:
        LOG.debug(f'Posting results to: {url}')
        LOG.debug(f'Posting results: {audit_report}')
        r = session.post(url=url, timeout=REQUEST_TIMEOUT, json=audit_report)
        LOG.debug(r.text)

        return r.json()
    except:
        LOG.exception("Error posting results")
        click.secho(
            "Warning: couldn't upload results to safetycli.com.",
            fg="yellow",
            file=sys.stderr
        )

        return {}


def fetch_database_file(path: str, db_name: str, cached = 0,
                        ecosystem: Optional[Ecosystem] = None):
    full_path = (Path(path) / (ecosystem.value if ecosystem else '') / db_name).expanduser().resolve()

    if not full_path.exists():
        raise DatabaseFileNotFoundError(db=path)

    with open(full_path) as f:
        data = json.loads(f.read())

    if cached:
        LOG.info('Writing %s to cache because cached value was %s', db_name, cached)
        write_to_cache(db_name, data)

    return data


def is_valid_database(db) -> bool:
    try:
        if db['meta']['schema_version'] == JSON_SCHEMA_VERSION:
            return True
    except Exception:
        return False

    return False


def fetch_database(session, full=False, db=False, cached=0, telemetry=True,
                   ecosystem: Optional[Ecosystem] = None, from_cache=True):

    if session.is_using_auth_credentials():
        mirrors = API_MIRRORS
    elif db:
        mirrors = [db]
    else:
        mirrors = OPEN_MIRRORS

    db_name = "insecure_full.json" if full else "insecure.json"
    for mirror in mirrors:
        # mirror can either be a local path or a URL
        if is_a_remote_mirror(mirror):
            if ecosystem is None:
                ecosystem = Ecosystem.PYTHON
            data = fetch_database_url(session, mirror, db_name=db_name, cached=cached,
                                      telemetry=telemetry, ecosystem=ecosystem, from_cache=from_cache)
        else:
            data = fetch_database_file(mirror, db_name=db_name, cached=cached,
                                       ecosystem=ecosystem)
        if data:
            if is_valid_database(data):
                return data
            raise MalformedDatabase(fetched_from=mirror,
                                    reason=f'Not supported schema version. '
                                           f'This Safety version supports only schema version {JSON_SCHEMA_VERSION}')

    raise DatabaseFetchError()


def get_vulnerabilities(pkg, spec, db):
    for entry in db['vulnerable_packages'][pkg]:
        for entry_spec in entry["specs"]:
            if entry_spec == spec:
                yield entry


def get_vulnerability_from(vuln_id, cve, data, specifier, db, name, pkg, ignore_vulns, affected):
    base_domain = db.get('meta', {}).get('base_domain')
    unpinned_ignored = ignore_vulns.get(vuln_id, {}).get('requirements', None)
    should_ignore = not unpinned_ignored or str(affected.specifier) in unpinned_ignored

    ignored = (ignore_vulns and vuln_id in ignore_vulns and should_ignore and (
            not ignore_vulns[vuln_id]['expires'] or ignore_vulns[vuln_id]['expires'] > datetime.utcnow()))
    more_info_url = f"{base_domain}{data.get('more_info_path', '')}"
    severity = None

    if cve and (cve.cvssv2 or cve.cvssv3):
        severity = Severity(source=cve.name, cvssv2=cve.cvssv2, cvssv3=cve.cvssv3)

    analyzed_requirement = affected
    analyzed_version = next(iter(analyzed_requirement.specifier)).version if is_pinned_requirement(
        analyzed_requirement.specifier) else None

    vulnerable_spec = set()
    vulnerable_spec.add(specifier)

    return Vulnerability(
        vulnerability_id=vuln_id,
        package_name=name,
        pkg=pkg,
        ignored=ignored,
        ignored_reason=ignore_vulns.get(vuln_id, {}).get('reason', None) if ignore_vulns else None,
        ignored_expires=ignore_vulns.get(vuln_id, {}).get('expires', None) if ignore_vulns else None,
        vulnerable_spec=vulnerable_spec,
        all_vulnerable_specs=data.get("specs", []),
        analyzed_version=analyzed_version,
        analyzed_requirement=analyzed_requirement,
        advisory=data.get("advisory"),
        is_transitive=data.get("transitive", False),
        published_date=data.get("published_date"),
        fixed_versions=[ver for ver in data.get("fixed_versions", []) if ver],
        closest_versions_without_known_vulnerabilities=data.get("closest_secure_versions", []),
        resources=data.get("vulnerability_resources"),
        CVE=cve,
        severity=severity,
        affected_versions=data.get("affected_versions", []),
        more_info_url=more_info_url
    )


def get_cve_from(data, db_full):
    try:
        xve_id: str = str(
            next(filter(lambda i: i.get('type', None) in ['cve', 'pve'], data.get('ids', []))).get('id', ''))
    except StopIteration:
        xve_id: str = ''

    if not xve_id:
        return None

    cve_meta = db_full.get("meta", {}).get("severities", {}).get(xve_id, {})
    return CVE(name=xve_id, cvssv2=cve_meta.get("cvssv2", None),
               cvssv3=cve_meta.get("cvssv3", None))


def ignore_vuln_if_needed(pkg: Package, vuln_id, cve, ignore_vulns, ignore_severity_rules, req):
    if not ignore_severity_rules:
        ignore_severity_rules = {}

    if not isinstance(ignore_vulns, dict):
        return

    severity = None

    if cve:
        if cve.cvssv2 and cve.cvssv2.get("base_score", None):
            severity = cve.cvssv2.get("base_score", None)

        if cve.cvssv3 and cve.cvssv3.get("base_score", None):
            severity = cve.cvssv3.get("base_score", None)

    ignore_severity_below = float(ignore_severity_rules.get('ignore-cvss-severity-below', 0.0))
    ignore_unknown_severity = bool(ignore_severity_rules.get('ignore-cvss-unknown-severity', False))

    if severity:
        if float(severity) < ignore_severity_below:
            reason = 'Ignored by severity rule in policy file, {0} < {1}'.format(float(severity), ignore_severity_below)
            ignore_vulns[vuln_id] = {'reason': reason, 'expires': None}
    elif ignore_unknown_severity:
        reason = 'Unknown CVSS severity, ignored by severity rule in policy file.'
        ignore_vulns[vuln_id] = {'reason': reason, 'expires': None}

    version = next(iter(req.specifier)).version if is_pinned_requirement(req.specifier) else pkg.version

    is_prev_not_ignored: bool = vuln_id not in ignore_vulns
    is_req_not_ignored: bool = 'requirements' in ignore_vulns.get(vuln_id, {}) and str(req.specifier) not in ignore_vulns.get(vuln_id, {}).get('requirements', set())

    if (is_prev_not_ignored or is_req_not_ignored) and is_ignore_unpinned_mode(version):
        reason = IGNORE_UNPINNED_REQ_REASON
        requirements = set()
        requirements.add(str(req.specifier))
        ignore_vulns[vuln_id] = {'reason': reason, 'expires': None, 'requirements': requirements}


def is_vulnerable(vulnerable_spec: SpecifierSet, requirement, package):

    if is_pinned_requirement(requirement.specifier):
        try:
            return vulnerable_spec.contains(next(iter(requirement.specifier)).version)
        except Exception:
            # Ugly for now...
            message = f'Version {requirement.specifier} for {package.name} is invalid and is ignored by Safety. Please See PEP 440.'
            if message not in [a['message'] for a in SafetyContext.local_announcements]:
                SafetyContext.local_announcements.append(
                    {'message': message,
                    'type': 'warning', 'local': True})
            return False

    return any(requirement.specifier.filter(vulnerable_spec.filter(package.insecure_versions, prereleases=True),
                                            prereleases=True))


@sync_safety_context
def check(*, session=None, packages=[], db_mirror=False, cached=0, ignore_vulns=None, ignore_severity_rules=None, proxy=None,
          include_ignored=False, is_env_scan=True, telemetry=True, params=None, project=None):
    SafetyContext().command = 'check'
    db = fetch_database(session, db=db_mirror, cached=cached, telemetry=telemetry)
    db_full = None
    vulnerable_packages = frozenset(db.get('vulnerable_packages', []))
    vulnerabilities = []
    found_pkgs = {}
    requirements = iter([])

    for p in packages:
        requirements = itertools.chain(p.requirements, requirements)
        found_pkgs[canonicalize_name(p.name)] = p

    # Let's report by req, pinned in environment will be ==version
    for req in requirements:
        vuln_per_req = {}
        name = canonicalize_name(req.name)

        pkg = found_pkgs.get(name, None)

        if not pkg.version:
            if not db_full:
                db_full = fetch_database(session, full=True, db=db_mirror, cached=cached,
                                         telemetry=telemetry)
            pkg.refresh_from(db_full)

        if name in vulnerable_packages:
            # we have a candidate here, build the spec set
            for specifier in db['vulnerable_packages'][name]:
                spec_set = SpecifierSet(specifiers=specifier)

                if is_vulnerable(spec_set, req, pkg):
                    if not db_full:
                        db_full = fetch_database(session, full=True, db=db_mirror, cached=cached,
                                                 telemetry=telemetry)
                    if not pkg.latest_version:
                        pkg.refresh_from(db_full)

                    for data in get_vulnerabilities(pkg=name, spec=specifier, db=db_full):
                        try:
                            vuln_id: str = str(next(filter(lambda i: i.get('type', None) == 'pyup', data.get('ids', []))).get('id', ''))
                        except StopIteration:
                            vuln_id: str = ''

                        if vuln_id in vuln_per_req:
                            vuln_per_req[vuln_id].vulnerable_spec.add(specifier)
                            continue

                        cve = get_cve_from(data, db_full)

                        ignore_vuln_if_needed(pkg, vuln_id, cve, ignore_vulns, ignore_severity_rules, req)

                        vulnerability = get_vulnerability_from(vuln_id, cve, data, specifier, db_full, name, pkg,
                                                               ignore_vulns, req)

                        should_add_vuln = not (vulnerability.is_transitive and is_env_scan)

                        if (include_ignored or vulnerability.vulnerability_id not in ignore_vulns) and should_add_vuln:
                            vuln_per_req[vulnerability.vulnerability_id] = vulnerability
                            vulnerabilities.append(vulnerability)

    return vulnerabilities, db_full


def precompute_remediations(remediations, packages, vulns, secure_vulns_by_user):

    for vuln in vulns:

        if vuln.ignored and vuln.ignored_reason != IGNORE_UNPINNED_REQ_REASON:
            secure_vulns_by_user.add(vuln.vulnerability_id)
            continue

        if vuln.package_name in remediations.keys() and str(vuln.analyzed_requirement.specifier) in remediations[vuln.package_name]:
            spec = remediations[vuln.package_name][str(vuln.analyzed_requirement.specifier)]
            spec['vulnerabilities_found'] = spec.get('vulnerabilities_found', 0) + 1
        else:
            version = None
            is_pinned = is_pinned_requirement(vuln.analyzed_requirement.specifier)

            if is_pinned:
                version = next(iter(vuln.analyzed_requirement.specifier)).version

            if not is_pinned and is_ignore_unpinned_mode(version):
                # Let's ignore this requirement
                continue

            vulns_count = 1
            packages[vuln.package_name] = vuln.pkg

            remediations[vuln.package_name][str(vuln.analyzed_requirement.specifier)] = {
                'vulnerabilities_found': vulns_count,
                'version': version,
                'requirement': vuln.analyzed_requirement,
                'more_info_url': vuln.pkg.more_info_url}


def get_closest_ver(versions, version, spec: SpecifierSet):
    results = {'upper': None, 'lower': None}

    if (not version and not spec) or not versions:
        return results

    sorted_versions = sorted(versions, key=lambda ver: parse_version(ver), reverse=True)

    if not version:
        sorted_versions = spec.filter(sorted_versions, prereleases=False)

        upper = None
        lower = None

        try:
            sorted_versions = list(sorted_versions)
            upper = sorted_versions[0]
            lower = sorted_versions[-1]
            results['upper'] = upper
            results['lower'] = lower if upper != lower else None
        except IndexError:
            pass

        return results

    current_v = parse_version(version)

    for v in sorted_versions:
        index = parse_version(v)

        if index > current_v:
            results['upper'] = index

        if index < current_v:
            results['lower'] = index
            break

    return results


def compute_sec_ver_for_user(package: Package, secure_vulns_by_user, db_full):
    versions = package.get_versions(db_full)
    affected_versions = []

    for vuln in db_full.get('vulnerable_packages', {}).get(package.name, []):
        vuln_id: str = str(next(filter(lambda i: i.get('type', None) == 'pyup', vuln.get('ids', []))).get('id', ''))
        if vuln_id and vuln_id not in secure_vulns_by_user:
            affected_versions += vuln.get('affected_versions', [])

    affected_v = set(affected_versions)
    sec_ver_for_user = list(versions.difference(affected_v))

    return sorted(sec_ver_for_user, key=lambda ver: parse_version(ver), reverse=True)


def compute_sec_ver(remediations, packages: Dict[str, Package], secure_vulns_by_user, db_full):
    """
    Compute the secure_versions and the closest_secure_version for each remediation using the affected_versions
    of each no ignored vulnerability of the same package, there is only a remediation for each package.
    """
    for pkg_name in remediations.keys():
        pkg: Package = packages.get(pkg_name, None)

        secure_versions = []

        if pkg:
            secure_versions = pkg.secure_versions

        analyzed = set(remediations[pkg_name].keys())

        if not is_using_api_key():
            continue

        for analyzed_requirement in analyzed:
            rem = remediations[pkg_name][analyzed_requirement]
            spec = rem.get('requirement').specifier
            version = rem['version']

            if not secure_vulns_by_user:
                secure_v = sorted(secure_versions, key=lambda ver: parse_version(ver), reverse=True)
            else:
                secure_v = compute_sec_ver_for_user(package=pkg, secure_vulns_by_user=secure_vulns_by_user, db_full=db_full)

            rem['closest_secure_version'] = get_closest_ver(secure_v, version, spec)

            upgrade = rem['closest_secure_version'].get('upper', None)
            downgrade = rem['closest_secure_version'].get('lower', None)
            recommended_version = None

            if upgrade:
                recommended_version = upgrade
            elif downgrade:
                recommended_version = downgrade

            rem['recommended_version'] = recommended_version
            rem['other_recommended_versions'] = [other_v for other_v in secure_v if
                                                                    other_v != str(recommended_version)]

            # Refresh the URL with the recommended version.

            spec = str(rem['requirement'].specifier)
            if is_using_api_key():
                rem['more_info_url'] = \
                    build_remediation_info_url(base_url=rem['more_info_url'], version=version,
                                           spec=spec,
                                           target_version=recommended_version)


def calculate_remediations(vulns, db_full):
    remediations = defaultdict(dict)
    package_metadata = {}
    secure_vulns_by_user = set()

    if not db_full:
        return remediations

    precompute_remediations(remediations, package_metadata, vulns, secure_vulns_by_user)
    compute_sec_ver(remediations, package_metadata, secure_vulns_by_user, db_full)

    return remediations


def should_apply_auto_fix(from_ver: Optional[Version], to_ver, allowed_automatic):
    if not from_ver:
        return False

    if 'major' in allowed_automatic:
        return True

    major_change = to_ver.major - from_ver.major
    minor_change = to_ver.minor - from_ver.minor

    if 'minor' in allowed_automatic:
        if major_change != 0:
            return False

        return True

    if 'patch' in allowed_automatic:
        if major_change != 0 or minor_change != 0:
            return False

        return True

    return False


def get_update_type(from_ver: Optional[Version], to_ver: Version):

    if not from_ver or (to_ver.major - from_ver.major) != 0:
        return 'major'

    if (to_ver.minor - from_ver.minor) != 0:
        return 'minor'

    return 'patch'


def process_fixes(files, remediations, auto_remediation_limit, output, no_output=True, prompt=False):
    req_remediations = itertools.chain.from_iterable(rem.values() for pkg_name, rem in remediations.items())
    requirements = compute_fixes_per_requirements(files, req_remediations, auto_remediation_limit, prompt=prompt)
    fixes = apply_fixes(requirements, output, no_output, prompt)
    return fixes


def process_fixes_scan(file_to_fix, to_fix_spec, auto_remediation_limit, output, no_output=True, prompt=False):
    to_fix_remediations =  []

    def get_remmediation_from(spec):
        upper = None
        lower = None
        recommended = None

        try:
            upper = Version(spec.remediation.closest_secure.upper) if spec.remediation.closest_secure.upper else None
        except Exception as e:
            LOG.error(f'Error getting upper remediation version, ignoring', exc_info=True)

        try:
            lower = Version(spec.remediation.closest_secure.lower) if spec.remediation.closest_secure.lower else None
        except Exception as e:
            LOG.error(f'Error getting lower remediation version, ignoring', exc_info=True)

        try:
            recommended = Version(spec.remediation.recommended)
        except Exception as e:
            LOG.error(f'Error getting recommended version for remediation, ignoring', exc_info=True)

        return {
            "vulnerabilities_found": spec.remediation.vulnerabilities_found,
            "version": next(iter(spec.specifier)).version if spec.is_pinned() else None,
            "requirement": spec,
            "more_info_url": spec.remediation.more_info_url,
            "closest_secure_version": {
                'upper': upper,
                'lower': lower
                },
            "recommended_version": recommended,
            "other_recommended_versions": spec.remediation.other_recommended
        }

    req_remediations = iter(get_remmediation_from(spec) for spec in to_fix_spec)
    SUPPORTED_FILE_TYPES = [FileType.REQUIREMENTS_TXT]

    if file_to_fix.file_type in SUPPORTED_FILE_TYPES:
        files = (open(file_to_fix.location),)
        requirements = compute_fixes_per_requirements(files, req_remediations, auto_remediation_limit, prompt=prompt)
    else:
        requirements = {
            'files': {str(file_to_fix.location): {'content': None, 'fixes': {'TO_SKIP': [], 'TO_APPLY': [], 'TO_CONFIRM': []}, 'supported': False, 'filename': file_to_fix.location.name}},
            'dependencies': defaultdict(dict),
        }

    fixes = apply_fixes(requirements, output, no_output, prompt, scan_flow=True, auto_remediation_limit=auto_remediation_limit)

    return fixes


def compute_fixes_per_requirements(files, req_remediations, auto_remediation_limit, prompt=False):
    requirements_files = get_requirements_content(files)

    from dparse.parser import parse, filetypes
    from packaging.version import Version, InvalidVersion

    requirements = {
        'files': {},
        'dependencies': defaultdict(dict),
    }

    for name, contents in requirements_files.items():
        dependency_file = parse(contents, path=name, file_type=filetypes.requirements_txt, resolve=True)
        dependency_files = dependency_file.resolved_files + [dependency_file]
        empty_spec = SpecifierSet('')
        default_spec = SpecifierSet('>=0')

        # Support recursive requirements in the multiple requirement files provided
        for resolved_f in dependency_files:
            if not resolved_f or isinstance(resolved_f, str):
                continue
            file = {'content': resolved_f.content, 'fixes': {'TO_SKIP': [], 'TO_APPLY': [], 'TO_CONFIRM': []}}
            requirements['files'][resolved_f.path] = file

            for d in resolved_f.dependencies:
                if d.specs == empty_spec:
                    d.specs = default_spec

                requirements['dependencies'][d.name][str(d.specs)] = (d, resolved_f.path)

    for remediation in req_remediations:
        req: SafetyRequirement = remediation.get('requirement')
        pkg: str = req.name

        dry_fix = Fix(package=pkg, more_info_url=remediation.get('more_info_url', ''),
                      previous_spec=str(req.specifier),
                      other_options=remediation.get('other_recommended_versions', []))
        from_ver: Optional[str] = remediation.get('version', None)

        if pkg not in requirements['dependencies'] or dry_fix.previous_spec not in requirements['dependencies'][pkg]:
            # Let's attach it to the first file scanned.
            file = next(iter(requirements['files']))
            # Let's use the no parsed version.
            dry_fix.previous_version = from_ver
            dry_fix.status = 'AUTOMATICALLY_SKIPPED_NOT_FOUND_IN_FILE'
            dry_fix.applied_at = file
            requirements['files'][file]['fixes']['TO_SKIP'].append(dry_fix)
            continue

        dependency, name = requirements['dependencies'][pkg][dry_fix.previous_spec]
        dry_fix.applied_at = name

        fixes = requirements['files'][name]['fixes']

        to_ver: Version = remediation['recommended_version']

        try:
            from_ver = parse_version(from_ver)
        except (InvalidVersion, TypeError):

            if not dry_fix.previous_spec:
                dry_fix.status = 'AUTOMATICALLY_SKIPPED_INVALID_VERSION'
                fixes['TO_SKIP'].append(dry_fix)
                continue

        dry_fix.previous_version = str(from_ver) if from_ver else from_ver

        if remediation['recommended_version'] is None:
            dry_fix.status = 'AUTOMATICALLY_SKIPPED_NO_RECOMMENDED_VERSION'
            fixes['TO_SKIP'].append(dry_fix)
            continue

        dry_fix.updated_version = str(to_ver)

        is_fixed = from_ver == to_ver

        if is_fixed:
            dry_fix.status = 'AUTOMATICALLY_SKIPPED_ALREADY_FIXED'
            fixes['TO_SKIP'].append(dry_fix)
            continue

        update_type = get_update_type(from_ver, to_ver)
        dry_fix.update_type = update_type
        dry_fix.dependency = dependency

        auto_fix = should_apply_auto_fix(from_ver, to_ver, auto_remediation_limit)

        TARGET = 'TO_APPLY'

        if auto_fix:
            dry_fix.status = 'PENDING_TO_APPLY'
            dry_fix.fix_type = 'AUTOMATIC'
        elif prompt:
            TARGET = 'TO_CONFIRM'
            dry_fix.status = 'PENDING_TO_CONFIRM'
            dry_fix.fix_type = 'MANUAL'
        else:
            TARGET = 'TO_SKIP'
            dry_fix.status = 'AUTOMATICALLY_SKIPPED_UNABLE_TO_CONFIRM'

        fixes[TARGET].append(dry_fix)

    return requirements


def apply_fixes(requirements, out_type, no_output, prompt, scan_flow=False, auto_remediation_limit=None):

    from dparse.updater import RequirementsTXTUpdater

    skip = []
    apply = []
    confirm = []

    brief = []

    if not no_output:
        style_kwargs = {}

        if not scan_flow:
            brief.append(('', {}))
            brief.append((f"Safety fix running", style_kwargs))
        print_service(brief, out_type)

    for name, data in requirements['files'].items():
        output = [('', {}),
                  (f"Analyzing {name}... [{get_fix_opt_used_msg(auto_remediation_limit)} limit]", {'styling': {'bold': True}, 'start_line_decorator': '->', 'indent': ' '})]

        r_skip = data['fixes']['TO_SKIP']
        r_apply = data['fixes']['TO_APPLY']
        r_confirm = data['fixes']['TO_CONFIRM']

        if data.get('supported', True):
            new_content = data['content']

            updated: bool = False

            for f in r_apply:
                new_content = RequirementsTXTUpdater.update(content=new_content, version=f.updated_version,
                                                            dependency=f.dependency, hashes=get_hashes(f.dependency))
                f.status = 'APPLIED'
                updated = True
                output.append(('', {}))
                output.append((f'- {get_applied_msg(f, mode="auto")}', {}))

            for f in r_skip:
                output.append(('', {}))
                output.append((f'- {get_skipped_msg(f)}', {}))

            if not no_output:
                print_service(output, out_type)

            if prompt and not no_output:
                for f in r_confirm:
                    options = [f"({index}) =={option}" for index, option in enumerate(f.other_options)]
                    input_hint = f'Enter “y” to update to {f.package}=={f.updated_version}, “n” to skip this package upgrade'

                    if len(options) > 0:
                        input_hint += f', or enter the index from these secure versions to upgrade to that version: {", ".join(options)}'

                    print_service([('', {})], out_type)
                    confirmed: str = prompt_service(
                        (f'- {f.package}{f.previous_spec} requires at least a {f.update_type} version update. Do you want to update {f.package} from {f.previous_spec} to =={f.updated_version}, which is the closest secure version? {input_hint}', {}),
                        out_type
                    ).lower()

                    try:
                        index: int = int(confirmed)
                        if index <= len(f.other_options):
                            confirmed = 'y'
                    except ValueError:
                        index = -1

                    if confirmed == 'y' or index > -1:
                        f.status = 'APPLIED'
                        updated = True

                        if index > -1:
                            f.updated_version = f.other_options[index]

                        new_content = RequirementsTXTUpdater.update(content=new_content, version=f.updated_version,
                                                                    dependency=f.dependency,
                                                                    hashes=get_hashes(f.dependency))
                        output.append((get_applied_msg(f, mode="manual"), {'indent': ' ' * 5}))
                    else:
                        f.status = 'MANUALLY_SKIPPED'
                        output.append((get_skipped_msg(f), {'indent': ' ' * 5}))

                    if not no_output:
                        print_service(output, out_type)

            if updated:
                output.append(('', {}))
                output.append((f"Updating {name}...", {}))
                with open(name, mode="w") as r_file:
                    r_file.write(new_content)
                output.append((f"Changes applied to {name}.", {}))
                count = len(r_apply) + len([1 for fix in r_confirm if fix.status == 'APPLIED'])
                output.append((f"{count} package {pluralize('version', count)} {pluralize('has', count)} been updated to secure versions in {Path(name).name}", {}))
                output.append(("Always check for breaking changes after updating packages.", {}))
            else:
                output.append((f"No fixes to be made in {name}.", {}))
                output.append(('', {}))
        else:
            not_supported_filename = data.get('filename', name)
            output.append(
                (f"{not_supported_filename} updates not supported: Please update these dependencies using your package manager.",
                 {'start_line_decorator': ' -', 'indent': ' '}))
            output.append(('', {}))

        if not no_output:
            print_service(output, out_type)

        skip.extend(r_skip)
        apply.extend(r_apply)
        confirm.extend(r_confirm)

    # The scan flow will handle the header and divider, because the scan flow can be called multiple times.
    if not no_output and not scan_flow:
        divider = f'{"=" * 78}' if out_type == 'text' else f'{"=" * (get_terminal_size().columns - 2)}'
        format_text = {'start_line_decorator': '+', 'end_line_decorator': '+', 'indent': ''}
        print_service([(divider, {})], out_type, format_text=format_text)

    return skip + apply + confirm


def find_vulnerabilities_fixed(vulnerabilities: Dict, fixes) -> List[Vulnerability]:
    fixed_specs = set(fix.previous_spec for fix in fixes)

    if not fixed_specs:
        return []

    return [vulnerability for vulnerability in vulnerabilities if
            str(vulnerability['analyzed_requirement'].specifier) in fixed_specs]


@sync_safety_context
def review(*, report=None, params=None):
    SafetyContext().command = 'review'
    vulnerable = []
    vulnerabilities = report.get('vulnerabilities', []) + report.get('ignored_vulnerabilities', [])
    remediations = defaultdict(dict)

    for key, pkg_rem in report.get('remediations', {}).items():
        analyzed = set(pkg_rem['requirements'].keys())

        for req in analyzed:
            req_rem = pkg_rem['requirements'][req]
            recommended = req_rem.get('recommended_version', None)
            secure_v = req_rem.get('other_recommended_versions', [])

            remediations[key][req] = {'vulnerabilities_found': req_rem.get('vulnerabilities_found', 0),
                                      'version': req_rem.get('version'),
                                      'requirement': SafetyRequirement(req_rem['requirement']['raw']),
                                      'other_recommended_versions': secure_v,
                                      'recommended_version': parse_version(recommended) if recommended else None,
                                      # minor isn't supported in review
                                      'more_info_url': req_rem.get('more_info_url')}

    packages = report.get('scanned_packages', [])
    pkgs = {}

    for name, values in packages.items():
        requirements = [SafetyRequirement(r['raw']) for r in values.get('requirements', [])]
        values.update({'requirements': requirements})
        pkgs[name] = Package(**values)

    ctx = SafetyContext()
    found_packages = list(pkgs.values())
    ctx.packages = found_packages
    ctx.review = report.get('report_meta', [])
    ctx.key = ctx.review.get('api_key', False)
    cvssv2 = None
    cvssv3 = None

    for vuln in vulnerabilities:
        vuln['pkg'] = pkgs.get(vuln.get('package_name', None))
        XVE_ID = vuln.get('CVE', None)  # Trying to get first the CVE ID

        severity = vuln.get('severity', None)
        if severity and severity.get('source', False):
            cvssv2 = severity.get('cvssv2', None)
            cvssv3 = severity.get('cvssv3', None)
            # Trying to get the PVE ID if it exists, otherwise it will be the same CVE ID of above
            XVE_ID = severity.get('source', False)
            vuln['severity'] = Severity(source=XVE_ID, cvssv2=cvssv2, cvssv3=cvssv3)
        else:
            vuln['severity'] = None

        ignored_expires = vuln.get('ignored_expires', None)

        if ignored_expires:
            vuln['ignored_expires'] = validate_expiration_date(ignored_expires)

        vuln['CVE'] = CVE(name=XVE_ID, cvssv2=cvssv2, cvssv3=cvssv3) if XVE_ID else None
        vuln['analyzed_requirement'] = SafetyRequirement(vuln['analyzed_requirement']['raw'])

        vulnerable.append(Vulnerability(**vuln))

    return vulnerable, remediations, found_packages


@sync_safety_context
def get_licenses(*, session=None, db_mirror=False, cached=0, telemetry=True):

    if db_mirror:
        mirrors = [db_mirror]
    else:
        mirrors = API_MIRRORS

    db_name = "licenses.json"

    for mirror in mirrors:
        # mirror can either be a local path or a URL
        if is_a_remote_mirror(mirror):
            licenses = fetch_database_url(session, mirror, db_name=db_name, cached=cached,
                                          telemetry=telemetry)
        else:
            licenses = fetch_database_file(mirror, db_name=db_name, ecosystem=None)
        if licenses:
            return licenses
    raise DatabaseFetchError()


def add_local_notifications(packages: List[Package],
                            ignore_unpinned_requirements: Optional[bool]) -> List[Dict[str, str]]:
    announcements = []
    unpinned_packages: [str] = [f"{pkg.name}" for pkg in packages if pkg.has_unpinned_req()]

    if unpinned_packages and ignore_unpinned_requirements is not False:
        found = len(unpinned_packages)
        and_msg = ''

        if found >= 2:
            last = unpinned_packages.pop()
            and_msg = f' and {last}'

        pkgs: str = f"{', '.join(unpinned_packages)}{and_msg} {'are' if found > 1 else 'is'}"
        doc_msg: str = get_specifier_range_info(style=False, pin_hint=True)

        if ignore_unpinned_requirements is None:
            msg = f'Warning: {pkgs} unpinned. Safety by default does not ' \
                  f'report on potential vulnerabilities in unpinned packages. {doc_msg}'
        else:
            msg = f'Warning: {pkgs} unpinned and potential vulnerabilities are ' \
                  f'being ignored given `ignore-unpinned-requirements` is True in your config. {doc_msg}'

        announcements.append({'message': msg, 'type': 'warning', 'local': True})

    announcements.extend(SafetyContext().local_announcements)

    return announcements


def get_announcements(session, telemetry=True, with_telemetry=None):
    LOG.info('Getting announcements')

    announcements = []

    url = f"{DATA_API_BASE_URL}announcements/"
    method = 'post'
    telemetry_data = with_telemetry if with_telemetry else build_telemetry_data(telemetry=telemetry)
    data = asdict(telemetry_data)
    request_kwargs = {'timeout': 3}
    data_keyword = 'json'

    source = os.environ.get('SAFETY_ANNOUNCEMENTS_URL', None)

    if source:
        LOG.debug(f'Getting the announcement from a different source: {source}')
        url = source
        method = 'get'
        data = {
            'telemetry': json.dumps(data)}
        data_keyword = 'params'

    request_kwargs[data_keyword] = data
    request_kwargs['url'] = url

    LOG.debug(f'Telemetry data sent: {data}')

    try:
        request_func = getattr(session, method)
        r = request_func(**request_kwargs)
        LOG.debug(r.text)
    except Exception as e:
        LOG.info('Unexpected but HANDLED Exception happened getting the announcements: %s', e)
        return announcements

    if r.status_code == 200:
        try:
            announcements = r.json()
            if 'announcements' in announcements.keys():
                announcements = announcements.get('announcements', [])
            else:
                LOG.info('There is not announcements key in the JSON response, is this a wrong structure?')
                announcements = []

        except json.JSONDecodeError as e:
            LOG.info('Unexpected but HANDLED Exception happened decoding the announcement response: %s', e)

    LOG.info('Announcements fetched')

    return announcements


def get_packages(files=False, stdin=False):

    if files:
        return list(itertools.chain.from_iterable(read_requirements(f, resolve=True) for f in files))

    if stdin:
        return list(read_requirements(sys.stdin))

    import pkg_resources

    def allowed_version(pkg: str, version: str):
        try:
            parse_version(version)
        except Exception:
            SafetyContext.local_announcements.append(
                {'message': f'Version {version} for {pkg} is invalid and is ignored by Safety. Please See PEP 440.',
                 'type': 'warning', 'local': True})
            return False

        return True

    w_set = pkg_resources.working_set

    SafetyContext().scanned_full_path.extend(w_set.entry_keys.keys())

    return [
        Package(name=d.key, version=d.version,
                requirements=[SafetyRequirement(f'{d.key}=={d.version}', found=d.location)],
                found=d.location, insecure_versions=[],
                secure_versions=[], latest_version=None, latest_version_without_known_vulnerabilities=None,
                more_info_url=None) for d in
        w_set
        if d.key not in {"python", "wsgiref", "argparse"} and allowed_version(d.key, d.version)
    ]


def read_vulnerabilities(fh):
    try:
        data = json.load(fh)
    except json.JSONDecodeError as e:
        raise MalformedDatabase(reason=e, fetched_from=fh.name)
    except TypeError as e:
        raise MalformedDatabase(reason=e, fetched_from=fh.name)

    return data


def get_server_policies(session, policy_file, proxy_dictionary: Dict):
    if session.api_key:
        server_policies = fetch_policy(session)
        server_audit_and_monitor = server_policies["audit_and_monitor"]
        server_safety_policy = server_policies["safety_policy"]
    else:
        server_audit_and_monitor = False
        server_safety_policy = ""

    if server_safety_policy and policy_file:
        click.secho(
            "Warning: both a local policy file '{policy_filename}' and a server sent policy are present. "
            "Continuing with the local policy file.".format(policy_filename=policy_file['filename']),
            fg="yellow",
            file=sys.stderr
        )
    elif server_safety_policy:
        with tempfile.NamedTemporaryFile(prefix='server-safety-policy-') as tmp:
            tmp.write(server_safety_policy.encode('utf-8'))
            tmp.seek(0)

            policy_file = SafetyPolicyFile().convert(tmp.name, param=None, ctx=None)
            LOG.info('Using server side policy file')

    return policy_file, server_audit_and_monitor


def save_report(path: str, default_name: str, report: str):
    if path:
        save_at = path

        if os.path.isdir(path):
            save_at = os.path.join(path, default_name)

        with open(save_at, 'w+') as report_file:
            report_file.write(report)



from abc import ABC, abstractmethod
from typing import List

from safety_schemas.models import Ecosystem, FileType, ConfigModel, \
    DependencyResultModel
from typer import FileTextWrite

NOT_IMPLEMENTED = "Not implemented funtion"


class Inspectable(ABC):

    @abstractmethod
    def inspect(self, config: ConfigModel) -> DependencyResultModel:
        return NotImplementedError(NOT_IMPLEMENTED)

    
class Remediable(ABC):

    @abstractmethod
    def remediate(self):
        return NotImplementedError(NOT_IMPLEMENTED)
    

class InspectableFile(Inspectable):
    
    def __init__(self, file: FileTextWrite):
        self.file = file
        self.ecosystem: Ecosystem
        self.file_type: FileType
        self.dependency_results: DependencyResultModel = \
            DependencyResultModel(dependencies=[])






from collections import defaultdict
from datetime import datetime
import itertools
import json
import logging
from pathlib import Path
import time
from typing import Any, Dict, List, Optional, Set
from rich.prompt import Prompt
from rich.text import Text
from rich.console import Console
from rich.padding import Padding
from safety_schemas.models import Vulnerability, ReportModel
import typer
from safety import safety
from safety.auth.constants import SAFETY_PLATFORM_URL
from safety.errors import SafetyException
from safety.output_utils import parse_html
from safety.scan.constants import DEFAULT_SPINNER

from safety_schemas.models import Ecosystem, FileType, PolicyFileModel, \
    PolicySource, ProjectModel, IgnoreCodes, Stage, PythonDependency

from safety.util import get_basic_announcements, get_safety_version

LOG = logging.getLogger(__name__)

import datetime

def render_header(targets: List[Path], is_system_scan: bool) -> Text:
    version = get_safety_version()
    scan_datetime = datetime.datetime.now(datetime.timezone.utc).strftime("%Y-%m-%d %H:%M:%S %Z")

    action = f"scanning {', '.join([str(t) for t in targets])}"
    if is_system_scan:
        action = "running [bold]system scan[/bold]"

    return Text.from_markup(
        f"[bold]Safety[/bold] {version} {action}\n{scan_datetime}")

def print_header(console, targets: List[Path], is_system_scan: bool = False):
    console.print(render_header(targets, is_system_scan), markup=True)

def print_announcements(console, ctx):    
    colors = {"error": "red", "warning": "yellow", "info": "default"}

    announcements = safety.get_announcements(ctx.obj.auth.client, 
                                             telemetry=ctx.obj.config.telemetry_enabled, 
                                             with_telemetry=ctx.obj.telemetry)
    basic_announcements = get_basic_announcements(announcements, False)

    if any(basic_announcements):
        console.print()
        console.print("[bold]Safety Announcements:[/bold]")
        console.print()
        for announcement in announcements:        
            color = colors.get(announcement.get('type', "info"), "default")
            console.print(f"[{color}]* {announcement.get('message')}[/{color}]")

def print_detected_ecosystems_section(console, file_paths: Dict[str, Set[Path]], 
                                      include_safety_prjs: bool = True):
    detected: Dict[Ecosystem, Dict[FileType, int]] = {}

    for file_type_key, f_paths in file_paths.items():
        file_type = FileType(file_type_key)
        if file_type.ecosystem:
            if file_type.ecosystem not in detected:
                detected[file_type.ecosystem] = {}
            detected[file_type.ecosystem][file_type] = len(f_paths)

    for ecosystem, f_type_count in detected.items():

        if not include_safety_prjs and ecosystem is Ecosystem.SAFETY_PROJECT:
            continue

        brief = "Found "
        file_types = []
        
        for f_type, count in f_type_count.items():
            file_types.append(f"{count} {f_type.human_name(plural=count>1)}")
        
        if len(file_types) > 1:
            brief += ", ".join(file_types[:-1]) + " and " + file_types[-1]
        else:
            brief += file_types[0]
        
        msg = f"{ecosystem.name.replace('_', ' ').title()} detected. {brief}"
        
        console.print(msg)

def print_brief(console, project: ProjectModel, dependencies_count: int = 0, 
                affected_count: int = 0, fixes_count: int = 0):
    from ..util import pluralize

    if project.policy:        
        if project.policy.source is PolicySource.cloud:
            policy_msg = f"policy fetched from Safety Platform"
        else:
            if project.id:
                policy_msg = f"local {project.id} project scan policy"
            else:
                policy_msg = f"local scan policy file"
    else:
        policy_msg = "default Safety CLI policies"

    console.print(f"Tested [number]{dependencies_count}[/number] {pluralize('dependency', dependencies_count)} for known security " \
                  f"issues using {policy_msg}")
    console.print(
        f"[number]{affected_count}[/number] security {pluralize('issue', affected_count)} found, [number]{fixes_count}[/number] {pluralize('fix', fixes_count)} suggested")
    
def print_fixes_section(console, requirements_txt_found: bool = False, is_detailed_output: bool = False):
    console.print("-" * console.size.width)
    console.print("Apply Fixes")
    console.print("-" * console.size.width)

    console.print()

    if requirements_txt_found:
        console.print("[green]Run `safety scan --apply-fixes`[/green] to update these packages and fix these vulnerabilities. "
                    "Documentation, limitations, and configurations for applying automated fixes: [link]https://docs.safetycli.com/safety-docs/vulnerability-remediation/applying-fixes[/link]")
        console.print()
        console.print("Alternatively, use your package manager to update packages to their secure versions. Always check for breaking changes when updating packages.")
    else:
        msg = "Use your package manager to update packages to their secure versions. Always check for breaking changes when updating packages."
        console.print(msg)

    if not is_detailed_output:
        console.print("[tip]Tip[/tip]: For more detailed output on each vulnerability, add the `--detailed-output` flag to safety scan.")

    console.print()
    console.print("-" * console.size.width)


def print_ignore_details(console, project: ProjectModel, ignored, 
                         is_detailed_output: bool = False, ignored_vulns_data = None):
    from ..util import pluralize

    if is_detailed_output:
        if not ignored_vulns_data:
            ignored_vulns_data = iter([])


        manual_ignored = {}
        cvss_severity_ignored = {}
        cvss_severity_ignored_pkgs = set()
        unpinned_ignored = {}
        unpinned_ignored_pkgs = set()
        environment_ignored = {}
        environment_ignored_pkgs = set()        

        for vuln_data in ignored_vulns_data:
            code = IgnoreCodes(vuln_data.ignored_code)
            if code is IgnoreCodes.manual:
                manual_ignored[vuln_data.vulnerability_id] = vuln_data
            elif code is IgnoreCodes.cvss_severity:
                cvss_severity_ignored[vuln_data.vulnerability_id] = vuln_data
                cvss_severity_ignored_pkgs.add(vuln_data.package_name)
            elif code is IgnoreCodes.unpinned_specification:
                unpinned_ignored[vuln_data.vulnerability_id] = vuln_data
                unpinned_ignored_pkgs.add(vuln_data.package_name)
            elif code is IgnoreCodes.environment_dependency:
                environment_ignored[vuln_data.vulnerability_id] = vuln_data
                environment_ignored_pkgs.add(vuln_data.package_name)                

        if manual_ignored:
            count = len(manual_ignored)
            console.print(
                f"[number]{count}[/number] were manually ignored due to the project policy:")
            for vuln in manual_ignored.values():
                render_to_console(vuln, console,
                                  rich_kwargs={"emoji": True, "overflow": "crop"}, 
                                  detailed_output=is_detailed_output)
        if cvss_severity_ignored:
            count = len(cvss_severity_ignored)
            console.print(
                f"[number]{count}[/number] {pluralize('vulnerability', count)} {pluralize('was', count)} ignored because " \
                     "of their severity or exploitability impacted the following" \
                        f" {pluralize('package', len(cvss_severity_ignored_pkgs))}: {', '.join(cvss_severity_ignored_pkgs)}"
            )
        if environment_ignored:
            count = len(environment_ignored)
            console.print(
                f"[number]{count}[/number] {pluralize('vulnerability', count)} {pluralize('was', count)} ignored because " \
                     "they are inside an environment dependency."
            )
        if unpinned_ignored:
            count = len(unpinned_ignored)
            console.print(
                f"[number]{count}[/number] {pluralize('vulnerability', count)} {pluralize('was', count)} ignored because " \
                    f"{pluralize('this', len(unpinned_ignored_pkgs))} {pluralize('package', len(unpinned_ignored_pkgs))} {pluralize('has', len(unpinned_ignored_pkgs))} unpinned specs: " \
                        f"{', '.join(unpinned_ignored_pkgs)}"
            )

    else:
        if len(ignored) > 0:
            console.print(f"([number]{len(ignored)}[/number] {pluralize('vulnerability', len(ignored))} {pluralize('was', len(ignored))} ignored due to " \
                          "project policy)")


def print_wait_project_verification(console, project_id, closure, on_error_delay=1):
    status = None
    wait_msg = f"Verifying project {project_id} with Safety Platform."

    with console.status(wait_msg, spinner=DEFAULT_SPINNER):
        try:
            f, kwargs = closure
            status = f(**kwargs)
        except Exception as e:
            LOG.exception(f'Unable to verify the project, reason: {e}')
            reason = "We are currently unable to verify the project, " \
                "and it is necessary to link the scan to a specific " \
                    f"project. Reason: {e}"
            raise SafetyException(message=reason)

        if not status:
            wait_msg = f'Unable to verify "{project_id}". Starting again...'
            time.sleep(on_error_delay)
    
    return status

def print_project_info(console, project: ProjectModel):
    config_msg = "loaded without policies or custom configuration."

    if project.policy:
        if project.policy.source is PolicySource.local:
            rel_location = project.policy.location.name if project.policy.location else ""
            config_msg = "configuration and policies fetched " \
                f"from {rel_location}."
        else:
            config_msg = " policies fetched " \
                "from Safety Platform."
    
    msg = f"[bold]{project.id} project found[/bold] - {config_msg}"
    console.print(msg)

def print_wait_policy_download(console, closure) -> Optional[PolicyFileModel]:
    policy = None
    wait_msg = "Looking for a policy from cloud..."

    with console.status(wait_msg, spinner=DEFAULT_SPINNER):
        try:
            f, kwargs = closure
            policy = f(**kwargs)
        except Exception as e:
            LOG.exception(f'Policy download failed, reason: {e}')
            console.print("Not using cloud policy file.")

        if policy:
            wait_msg = "Policy fetched from Safety Platform."
        else:
            # TODO: Send a log
            pass
    return policy


def prompt_project_id(console, stage: Stage, 
                      prj_root_name: Optional[str],
                      do_not_exit=True) -> str:
    from safety.util import clean_project_id
    default_prj_id = clean_project_id(prj_root_name) if prj_root_name else None

    non_interactive_mode = console.quiet or not console.is_interactive
    if stage is not Stage.development and non_interactive_mode:
        # Fail here
        console.print("The scan needs to be linked to a project.")
        raise typer.Exit(code=1)
    
    hint = ""
    if default_prj_id:        
        hint = f" If empty Safety will use [bold]{default_prj_id}[/bold]"    
    prompt_text = f"Set a project id for this scan (no spaces).{hint}"

    def ask():
        prj_id = None

        result = Prompt.ask(prompt_text, default=None, console=console)

        if result:
            prj_id = clean_project_id(result)
        elif default_prj_id:
            prj_id = default_prj_id

        return prj_id

    project_id = ask()

    while not project_id and do_not_exit:
        project_id = ask()

    return project_id


def prompt_link_project(console, prj_name: str, prj_admin_email: str) -> bool:
    console.print("[bold]Safety found an existing project with this name in your organization:[/bold]")

    for detail in (f"[bold]Project name:[/bold] {prj_name}", 
                   f"[bold]Project admin:[/bold] {prj_admin_email}"):
        console.print(Padding(detail, (0, 0, 0, 2)), emoji=True)

    prompt_question = "Do you want to link this scan with this existing project?"
    
    answer = Prompt.ask(prompt=prompt_question, choices=["y", "n"], 
                        default="y", show_default=True, console=console).lower()
    
    return answer == "y"


def render_to_console(cls: Vulnerability, console: Console, rich_kwargs,
                      detailed_output: bool = False):
    cls.__render__(console, detailed_output, rich_kwargs)


def get_render_console(entity_type):

    if entity_type is Vulnerability:
        def __render__(self, console: Console, detailed_output: bool, rich_kwargs):
            if not rich_kwargs:
                rich_kwargs = {}

            pre = " Ignored:" if self.ignored else ""
            severity_detail = None

            if self.severity and self.severity.source:
                severity_detail = self.severity.source

                if self.severity.cvssv3 and "base_severity" in self.severity.cvssv3:
                    severity_detail += f", CVSS Severity {self.severity.cvssv3['base_severity'].upper()}"

            advisory_length = 200 if detailed_output else 110

            console.print(
                Padding(
                    f"->{pre} Vuln ID [vuln_id]{self.vulnerability_id}[/vuln_id]: {severity_detail if severity_detail else ''}", 
                    (0, 0, 0, 2)
                ), **rich_kwargs)
            console.print(
                Padding(
                    f"{self.advisory[:advisory_length]}{'...' if len(self.advisory) > advisory_length else ''}", 
                    (0, 0, 0, 5)
                ), **rich_kwargs)

            if detailed_output:
                console.print(
                    Padding(f"For more information: [link]{self.more_info_url}[/link]", (0, 0, 0, 5)),
                      **rich_kwargs)

        return __render__


def render_scan_html(report: ReportModel, obj) -> str:
    from safety.scan.command import ScannableEcosystems

    project = report.projects[0] if any(report.projects) else None

    scanned_packages = 0
    affected_packages = 0
    ignored_packages = 0
    remediations_recommended = 0
    ignored_vulnerabilities = 0
    vulnerabilities = 0
    vulns_per_file = defaultdict(int)
    remed_per_file = defaultdict(int)

    for file in project.files:
        scanned_packages += len(file.results.dependencies)
        affected_packages += len(file.results.get_affected_dependencies())
        ignored_vulnerabilities += len(file.results.ignored_vulns)

        for spec in file.results.get_affected_specifications():
            vulnerabilities += len(spec.vulnerabilities)
            vulns_per_file[file.location] += len(spec.vulnerabilities)
            if spec.remediation:
                remed_per_file[file.location] += 1
                remediations_recommended += 1

        ignored_packages += len(file.results.ignored_vulns)

    # TODO: Get this information for the report model (?)
    summary = {"scanned_packages": scanned_packages, 
               "affected_packages": affected_packages, 
               "remediations_recommended": remediations_recommended,
               "ignored_vulnerabilities": ignored_vulnerabilities, "vulnerabilities": vulnerabilities}
    
    vulnerabilities = []
    
    
    # TODO: This should be based on the configs per command
    ecosystems = [(f"{ecosystem.name.title()}",
                  [file_type.human_name(plural=True) for file_type in ecosystem.file_types]) for ecosystem in [Ecosystem(member.value) for member in list(ScannableEcosystems)]]
    
    settings ={"audit_and_monitor": True, "platform_url": SAFETY_PLATFORM_URL, "ecosystems": ecosystems}
    template_context = {"report": report, "summary": summary, "announcements": [], 
                        "project": project, 
                        "platform_enabled": obj.platform_enabled,
                        "settings": settings,
                        "vulns_per_file": vulns_per_file,
                        "remed_per_file": remed_per_file}
    
    return parse_html(kwargs=template_context, template="scan/index.html")


def generate_spdx_creation_info(*, spdx_version: str, project_identifier: str) -> Any:
    from spdx_tools.spdx.model import (
        Actor,
        ActorType,
        CreationInfo,
    )

    version = int(time.time())
    SPDX_ID_TYPE = "SPDXRef-DOCUMENT"
    DOC_NAME = f"{project_identifier}-{version}"

    DOC_NAMESPACE = f"https://spdx.org/spdxdocs/{DOC_NAME}"
    # DOC_NAMESPACE = f"urn:safety:{project_identifier}:{version}"

    DOC_COMMENT = f"This document was created using SPDX {spdx_version}"
    CREATOR_COMMENT = "Safety CLI automatically created this SPDX document from a scan report."

    from ..util import get_safety_version
    TOOL_ID = "safety"
    TOOL_VERSION = get_safety_version()

    doc_creator = Actor(
        actor_type=ActorType.TOOL,
        name=f"{TOOL_ID}-{TOOL_VERSION}",
        email=None
    )

    creation_info = CreationInfo(
        spdx_version=f"SPDX-{spdx_version}",
        spdx_id=SPDX_ID_TYPE,
        name=DOC_NAME,
        document_namespace=DOC_NAMESPACE,
        creators=[doc_creator],
        created=datetime.datetime.now(),
        document_comment=DOC_COMMENT,
        creator_comment=CREATOR_COMMENT
    )
    return creation_info


def create_pkg_ext_ref(*, package: PythonDependency, version: Optional[str]):
    from spdx_tools.spdx.model import (
        ExternalPackageRef,
        ExternalPackageRefCategory,
    )

    version_detail = f'@{version}' if version else ''
    pkg_ref = ExternalPackageRef(
        ExternalPackageRefCategory.PACKAGE_MANAGER,
        "purl",
        f"pkg:pypi/{package.name}{version_detail}",
    )
    return pkg_ref


def create_packages(dependencies: List[PythonDependency]) -> List[Any]:
    from spdx_tools.spdx.model.spdx_no_assertion import SpdxNoAssertion

    from spdx_tools.spdx.model import (
        Package,
    )    

    doc_pkgs = []
    pkgs_added = set([])
    for dependency in dependencies:
        for spec in dependency.specifications:
            pkg_version = next(iter(spec.specifier)).version if spec.is_pinned() else f"{spec.specifier}"
            dep_name = dependency.name.replace('_', '-')
            pkg_id = f"SPDXRef-pip-{dep_name}-{pkg_version}" if spec.is_pinned() else f"SPDXRef-pip-{dep_name}"
            if pkg_id in pkgs_added:
                continue
            pkg_ref = create_pkg_ext_ref(package=dependency, version=pkg_version)
            
            pkg = Package(
                spdx_id=pkg_id,
                name=f"pip:{dep_name}",
                download_location=SpdxNoAssertion(),
                version=pkg_version,
                file_name="",
                supplier=SpdxNoAssertion(),
                files_analyzed=False,
                license_concluded=SpdxNoAssertion(),
                license_declared=SpdxNoAssertion(),
                copyright_text=SpdxNoAssertion(),
                external_references=[pkg_ref],
            )
            pkgs_added.add(pkg_id)
            doc_pkgs.append(pkg)
    return doc_pkgs


def create_spdx_document(*, report: ReportModel, spdx_version: str) -> Optional[Any]:
    from spdx_tools.spdx.model import (
        Document,
        Relationship,
        RelationshipType,
    )

    project =  report.projects[0] if any(report.projects) else None

    if not project:
        return None
    
    prj_id = project.id
    
    if not prj_id:
        parent_name = project.project_path.parent.name
        prj_id = parent_name if parent_name else str(int(time.time()))
    
    creation_info = generate_spdx_creation_info(spdx_version=spdx_version, project_identifier=prj_id)

    depedencies = iter([])
    for file in project.files:
        depedencies = itertools.chain(depedencies, file.results.dependencies)

    packages = create_packages(depedencies)

    # Requirement for document to have atleast one relationship
    relationship = Relationship(
        "SPDXRef-DOCUMENT",
        RelationshipType.DESCRIBES,
        "SPDXRef-DOCUMENT"
    )
    spdx_doc = Document(
        creation_info,
        packages,
        [],
        [],
        [],
        [relationship],
        []
    )
    return spdx_doc


def render_scan_spdx(report: ReportModel, obj, spdx_version: Optional[str]) -> Optional[Any]:
    from spdx_tools.spdx.writer.write_utils import (
        convert,
        validate_and_deduplicate
    )
        
    # Set to latest supported if a version is not specified
    if not spdx_version:
        spdx_version = "2.3"

    document_obj = create_spdx_document(report=report, spdx_version=spdx_version)
    document_obj = validate_and_deduplicate(document=document_obj, validate=True, drop_duplicates=True)
    doc = None

    if document_obj:
        doc = convert(document=document_obj, converter=None)

    return json.dumps(doc) if doc else None



import logging
import os
import platform
import re
import sys
from collections import defaultdict
from datetime import datetime
from difflib import SequenceMatcher
from threading import Lock
from typing import List, Optional

import click
from click import BadParameter
from dparse import parse, filetypes
from packaging.utils import canonicalize_name
from packaging.version import parse as parse_version
from packaging.specifiers import SpecifierSet
from requests import PreparedRequest
from ruamel.yaml import YAML
from ruamel.yaml.error import MarkedYAMLError

from safety.constants import EXIT_CODE_FAILURE, EXIT_CODE_OK, HASH_REGEX_GROUPS, SYSTEM_CONFIG_DIR, USER_CONFIG_DIR
from safety.errors import InvalidProvidedReportError
from safety.models import Package, RequirementFile, is_pinned_requirement, SafetyRequirement
from safety_schemas.models import TelemetryModel

LOG = logging.getLogger(__name__)


def is_a_remote_mirror(mirror):
    return mirror.startswith("http://") or mirror.startswith("https://")


def is_supported_by_parser(path):
    supported_types = (".txt", ".in", ".yml", ".ini", "Pipfile",
                       "Pipfile.lock", "setup.cfg", "poetry.lock")
    return path.endswith(supported_types)


def parse_requirement(dep, found):
    req = SafetyRequirement(dep)
    req.found = found

    if req.specifier == SpecifierSet(''):
        req.specifier = SpecifierSet('>=0')

    return req


def find_version(requirements):
    ver = None

    if len(requirements) != 1:
        return ver

    specs = requirements[0].specifier

    if is_pinned_requirement(specs):
        ver = next(iter(requirements[0].specifier)).version

    return ver


def read_requirements(fh, resolve=True):
    """
    Reads requirements from a file like object and (optionally) from referenced files.
    :param fh: file like object to read from
    :param resolve: boolean. resolves referenced files.
    :return: generator
    """
    is_temp_file = not hasattr(fh, 'name')
    path = None
    found = 'temp_file'
    file_type = filetypes.requirements_txt
    absolute_path: Optional[str] = None

    if not is_temp_file and is_supported_by_parser(fh.name):
        LOG.debug('not temp and a compatible file')
        path = fh.name
        absolute_path = os.path.abspath(path)
        SafetyContext().scanned_full_path.append(absolute_path)
        found = path
        file_type = None

    LOG.debug(f'Path: {path}')
    LOG.debug(f'File Type: {file_type}')
    LOG.debug('Trying to parse file using dparse...')
    content = fh.read()
    LOG.debug(f'Content: {content}')
    dependency_file = parse(content, path=path, resolve=resolve,
                            file_type=file_type)
    LOG.debug(f'Dependency file: {dependency_file.serialize()}')
    LOG.debug(f'Parsed, dependencies: {[dep.serialize() for dep in dependency_file.resolved_dependencies]}')

    reqs_pkg = defaultdict(list)

    for req in dependency_file.resolved_dependencies:
        reqs_pkg[canonicalize_name(req.name)].append(req)

    for pkg, reqs in reqs_pkg.items():
        requirements = list(map(lambda req: parse_requirement(req, absolute_path), reqs))
        version = find_version(requirements)

        yield Package(name=pkg, version=version,
                      requirements=requirements,
                      found=found,
                      absolute_path=absolute_path,
                      insecure_versions=[],
                      secure_versions=[], latest_version=None,
                      latest_version_without_known_vulnerabilities=None,
                      more_info_url=None)


def get_proxy_dict(proxy_protocol, proxy_host, proxy_port):
    if proxy_protocol and proxy_host and proxy_port:
        # Safety only uses https request, so only https dict will be passed to requests
        return {'https': f"{proxy_protocol}://{proxy_host}:{str(proxy_port)}"}
    return None


def get_license_name_by_id(license_id, db):
    licenses = db.get('licenses', [])
    for name, id in licenses.items():
        if id == license_id:
            return name
    return None


def get_flags_from_context():
    flags = {}
    context = click.get_current_context(silent=True)

    if context:
        for option in context.command.params:
            flags_per_opt = option.opts + option.secondary_opts
            for flag in flags_per_opt:
                flags[flag] = option.name

    return flags


def get_used_options():
    flags = get_flags_from_context()
    used_options = {}

    for arg in sys.argv:
        cleaned_arg = arg if '=' not in arg else arg.split('=')[0]
        if cleaned_arg in flags:
            option_used = flags.get(cleaned_arg)

            if option_used in used_options:
                used_options[option_used][cleaned_arg] = used_options[option_used].get(cleaned_arg, 0) + 1
            else:
                used_options[option_used] = {cleaned_arg: 1}

    return used_options


def get_safety_version():
    from safety import VERSION
    return VERSION


def get_primary_announcement(announcements):
    for announcement in announcements:
        if announcement.get('type', '').lower() == 'primary_announcement':
            try:
                from safety.output_utils import build_primary_announcement
                build_primary_announcement(announcement, columns=80)
            except Exception as e:
                LOG.debug(f'Failed to build primary announcement: {str(e)}')
                return None

            return announcement

    return None


def get_basic_announcements(announcements, include_local: bool = True):
    return [announcement for announcement in announcements if
            announcement.get('type', '').lower() != 'primary_announcement' and not announcement.get('local', False)
            or (announcement.get('local', False) and include_local)]


def filter_announcements(announcements, by_type='error'):
    return [announcement for announcement in announcements if
            announcement.get('type', '').lower() == by_type]


def build_telemetry_data(telemetry = True, 
                         command: Optional[str] = None, 
                         subcommand: Optional[str] = None) -> TelemetryModel:
    context = SafetyContext()

    body = {
        'os_type': os.environ.get("SAFETY_OS_TYPE", None) or platform.system(),
        'os_release': os.environ.get("SAFETY_OS_RELEASE", None) or platform.release(),
        'os_description': os.environ.get("SAFETY_OS_DESCRIPTION", None) or platform.platform(),
        'python_version': platform.python_version(),
        'safety_command': command if command else context.command,
        'safety_options': get_used_options()
    } if telemetry else {}

    body['safety_version'] = get_safety_version()
    body['safety_source'] = os.environ.get("SAFETY_SOURCE", None) or context.safety_source

    if not 'safety_options' in body:
        body['safety_options'] = {}

    LOG.debug(f'Telemetry body built: {body}')

    return TelemetryModel(**body)


def build_git_data():
    import subprocess

    def git_command(commandline):
        return subprocess.run(commandline, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL).stdout.decode('utf-8').strip()

    try:
        is_git = git_command(["git", "rev-parse", "--is-inside-work-tree"])
    except Exception:
        is_git = False

    if is_git == "true":
        result = {
            "branch": "",
            "tag": "",
            "commit": "",
            "dirty": "",
            "origin": ""
        }

        try:
            result['branch'] = git_command(["git", "symbolic-ref", "--short", "-q", "HEAD"])
            result['tag'] = git_command(["git", "describe", "--tags", "--exact-match"])

            commit = git_command(["git", "describe", '--match=""', '--always', '--abbrev=40', '--dirty'])
            result['dirty'] = commit.endswith('-dirty')
            result['commit'] = commit.split("-dirty")[0]

            result['origin'] = git_command(["git", "remote", "get-url", "origin"])
        except Exception:
            pass

        return result
    else:
        return {
            "error": "not-git-repo"
        }


def output_exception(exception, exit_code_output=True):
    click.secho(str(exception), fg="red", file=sys.stderr)

    if exit_code_output:
        exit_code = EXIT_CODE_FAILURE
        if hasattr(exception, 'get_exit_code'):
            exit_code = exception.get_exit_code()
    else:
        exit_code = EXIT_CODE_OK

    sys.exit(exit_code)

def build_remediation_info_url(base_url: str, version: Optional[str], spec: str,
                               target_version: Optional[str] = ''):

    params = {'from': version, 'to': target_version}

    # No pinned version
    if not version:
        params = {'spec': spec}

    req = PreparedRequest()
    req.prepare_url(base_url, params)

    return req.url

def get_processed_options(policy_file, ignore, ignore_severity_rules, exit_code, ignore_unpinned_requirements=None,
                          project=None):
    if policy_file:
        project_config = policy_file.get('project', {})
        security = policy_file.get('security', {})
        ctx = click.get_current_context()
        source = ctx.get_parameter_source("exit_code")

        if not project:
            project_id = project_config.get('id', None)
            if not project_id:
                project_id = None
            project = project_id

        if ctx.get_parameter_source("ignore_unpinned_requirements") == click.core.ParameterSource.DEFAULT:
            ignore_unpinned_requirements = security.get('ignore-unpinned-requirements', None)

        if not ignore:
            ignore = security.get('ignore-vulnerabilities', {})
        if source == click.core.ParameterSource.DEFAULT:
            exit_code = not security.get('continue-on-vulnerability-error', False)
        ignore_cvss_below = security.get('ignore-cvss-severity-below', 0.0)
        ignore_cvss_unknown = security.get('ignore-cvss-unknown-severity', False)
        ignore_severity_rules = {'ignore-cvss-severity-below': ignore_cvss_below,
                                 'ignore-cvss-unknown-severity': ignore_cvss_unknown}

    return ignore, ignore_severity_rules, exit_code, ignore_unpinned_requirements, project


def get_fix_options(policy_file, auto_remediation_limit):
    auto_fix = []

    source = click.get_current_context().get_parameter_source("auto_remediation_limit")
    if source == click.core.ParameterSource.COMMANDLINE:
        return auto_remediation_limit

    if policy_file:
        fix = policy_file.get('security-updates', {})
        auto_fix = fix.get('auto-security-updates-limit', None)
        if not auto_fix:
            auto_fix = []

    return auto_fix


class MutuallyExclusiveOption(click.Option):
    def __init__(self, *args, **kwargs):
        self.mutually_exclusive = set(kwargs.pop('mutually_exclusive', []))
        self.with_values = kwargs.pop('with_values', {})
        help = kwargs.get('help', '')
        if self.mutually_exclusive:
            ex_str = ', '.join(["{0} with values {1}".format(item, self.with_values.get(item)) if item in self.with_values else item for item in self.mutually_exclusive])
            kwargs['help'] = help + (
                ' NOTE: This argument is mutually exclusive with '
                ' arguments: [' + ex_str + '].'
            )
        super(MutuallyExclusiveOption, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        m_exclusive_used = self.mutually_exclusive.intersection(opts)
        option_used = m_exclusive_used and self.name in opts

        exclusive_value_used = False
        for used in m_exclusive_used:
            value_used = opts.get(used, None)
            if not isinstance(value_used, List):
                value_used = [value_used]
            if value_used and set(self.with_values.get(used, [])).intersection(value_used):
                exclusive_value_used = True

        if option_used and (not self.with_values or exclusive_value_used):
            options = ', '.join(self.opts)
            prohibited = ''.join(["\n * --{0} with {1}".format(item, self.with_values.get(
                        item)) if item in self.with_values else f"\n * {item}" for item in self.mutually_exclusive])
            raise click.UsageError(
                f"Illegal usage: `{options}` is mutually exclusive with: {prohibited}"
            )

        return super(MutuallyExclusiveOption, self).handle_parse_result(
            ctx,
            opts,
            args
        )


class DependentOption(click.Option):
    def __init__(self, *args, **kwargs):
        self.required_options = set(kwargs.pop('required_options', []))
        help = kwargs.get('help', '')
        if self.required_options:
            ex_str = ', '.join(self.required_options)
            kwargs['help'] = help + (
                f" Requires: [ {ex_str} ]"
            )
        super(DependentOption, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        missing_required_arguments = None

        if self.name in opts:
            missing_required_arguments = self.required_options.difference(opts)

        if missing_required_arguments:
            raise click.UsageError(
                "Illegal usage: `{}` needs the "
                "arguments `{}`.".format(
                    self.name,
                    ', '.join(missing_required_arguments)
                )
            )

        return super(DependentOption, self).handle_parse_result(
            ctx,
            opts,
            args
        )


def transform_ignore(ctx, param, value):
    ignored_default_dict = {'reason': '', 'expires': None}
    if isinstance(value, tuple) and any(value):
        # Following code is required to support the 2 ways of providing 'ignore'
        # --ignore=1234,567,789
        # or, the historical way (supported for backward compatibility)
        # -i 1234 -i 567
        combined_value = ','.join(value)
        ignore_ids = {vuln_id.strip() for vuln_id in combined_value.split(',')}
        return {ignore_id: dict(ignored_default_dict) for ignore_id in ignore_ids}

    return {}


def active_color_if_needed(ctx, param, value):
    if value == 'screen':
        ctx.color = True

    color = os.environ.get("SAFETY_COLOR", None)

    if color is not None:
        color = color.lower()

        if color == '1' or color == 'true':
            ctx.color = True
        elif color == '0' or color == 'false':
            ctx.color = False

    return value


def json_alias(ctx, param, value):
    if value:
        os.environ['SAFETY_OUTPUT'] = 'json'
        return value

def html_alias(ctx, param, value):
    if value:
        os.environ['SAFETY_OUTPUT'] = 'html'
        return value


def bare_alias(ctx, param, value):
    if value:
        os.environ['SAFETY_OUTPUT'] = 'bare'
        return value


def get_terminal_size():
    from shutil import get_terminal_size as t_size
    # get_terminal_size can report 0, 0 if run from pseudo-terminal prior Python 3.11 versions

    columns = t_size().columns or 80
    lines = t_size().lines or 24

    return os.terminal_size((columns, lines))


def clean_project_id(input_string):
    input_string = re.sub(r'[^a-zA-Z0-9]+', '-', input_string)
    input_string = input_string.strip('-')
    input_string = input_string.lower()

    return input_string


def validate_expiration_date(expiration_date):
    d = None

    if expiration_date:
        try:
            d = datetime.strptime(expiration_date, '%Y-%m-%d')
        except ValueError as e:
            pass

        try:
            d = datetime.strptime(expiration_date, '%Y-%m-%d %H:%M:%S')
        except ValueError as e:
            pass

    return d


class SafetyPolicyFile(click.ParamType):
    """
       Custom Safety Policy file to hold validations
    """

    name = "filename"
    envvar_list_splitter = os.path.pathsep

    def __init__(
        self,
        mode: str = "r",
        encoding: str = None,
        errors: str = "strict",
        pure: bool = os.environ.get('SAFETY_PURE_YAML', 'false').lower() == 'true'
    ) -> None:
        self.mode = mode
        self.encoding = encoding
        self.errors = errors
        self.basic_msg = '\n' + click.style('Unable to load the Safety Policy file "{name}".', fg='red')
        self.pure = pure

    def to_info_dict(self):
        info_dict = super().to_info_dict()
        info_dict.update(mode=self.mode, encoding=self.encoding)
        return info_dict

    def fail_if_unrecognized_keys(self, used_keys, valid_keys, param=None, ctx=None, msg='{hint}', context_hint=''):
        for keyword in used_keys:
            if keyword not in valid_keys:
                match = None
                max_ratio = 0.0
                if isinstance(keyword, str):
                    for option in valid_keys:
                        ratio = SequenceMatcher(None, keyword, option).ratio()
                        if ratio > max_ratio:
                            match = option
                            max_ratio = ratio

                maybe_msg = f' Maybe you meant: {match}' if max_ratio > 0.7 else \
                            f' Valid keywords in this level are: {", ".join(valid_keys)}'

                self.fail(msg.format(hint=f'{context_hint}"{keyword}" is not a valid keyword.{maybe_msg}'), param, ctx)

    def fail_if_wrong_bool_value(self, keyword, value, msg='{hint}'):
        if value is not None and not isinstance(value, bool):
            self.fail(msg.format(hint=f"'{keyword}' value needs to be a boolean. "
                                      "You can use True, False, TRUE, FALSE, true or false"))

    def convert(self, value, param, ctx):
        try:

            if hasattr(value, "read") or hasattr(value, "write"):
                return value

            msg = self.basic_msg.format(name=value) + '\n' + click.style('HINT:', fg='yellow') + ' {hint}'

            f, _ = click.types.open_stream(
                value, self.mode, self.encoding, self.errors, atomic=False
            )
            filename = ''

            try:
                raw = f.read()
                yaml = YAML(typ='safe', pure=self.pure)
                safety_policy = yaml.load(raw)
                filename = f.name
                f.close()
            except Exception as e:
                show_parsed_hint = isinstance(e, MarkedYAMLError)
                hint = str(e)
                if show_parsed_hint:
                    hint = f'{str(e.problem).strip()} {str(e.context).strip()} {str(e.context_mark).strip()}'

                self.fail(msg.format(name=value, hint=hint), param, ctx)

            if not safety_policy or not isinstance(safety_policy, dict) or not safety_policy.get('security', None):
                hint = "you are missing the security root tag"
                try:
                    version = safety_policy["version"]
                    if version:
                        hint = f"{filename} is a policy file version {version}. " \
                            "Legacy policy file parser only accepts versions minor than 3.0" \
                            "\nNote: `safety check` command accepts policy file versions <= 2.0. Versions >= 2.0 are not supported."
                except Exception:
                    pass
                self.fail(
                    msg.format(hint=hint), param, ctx)

            security_config = safety_policy.get('security', {})
            security_keys = ['ignore-cvss-severity-below', 'ignore-cvss-unknown-severity', 'ignore-vulnerabilities',
                             'continue-on-vulnerability-error', 'ignore-unpinned-requirements']
            self.fail_if_unrecognized_keys(security_config.keys(), security_keys, param=param, ctx=ctx, msg=msg,
                                           context_hint='"security" -> ')

            ignore_cvss_security_below = security_config.get('ignore-cvss-severity-below', None)

            if ignore_cvss_security_below:
                limit = 0.0

                try:
                    limit = float(ignore_cvss_security_below)
                except ValueError as e:
                    self.fail(msg.format(hint="'ignore-cvss-severity-below' value needs to be an integer or float."))

                if limit < 0 or limit > 10:
                    self.fail(msg.format(hint="'ignore-cvss-severity-below' needs to be a value between 0 and 10"))

            continue_on_vulnerability_error = security_config.get('continue-on-vulnerability-error', None)
            self.fail_if_wrong_bool_value('continue-on-vulnerability-error', continue_on_vulnerability_error, msg)

            ignore_cvss_unknown_severity = security_config.get('ignore-cvss-unknown-severity', None)
            self.fail_if_wrong_bool_value('ignore-cvss-unknown-severity', ignore_cvss_unknown_severity, msg)

            ignore_vulns = safety_policy.get('security', {}).get('ignore-vulnerabilities', {})

            if ignore_vulns:
                if not isinstance(ignore_vulns, dict):
                    self.fail(msg.format(hint="Vulnerability IDs under the 'ignore-vulnerabilities' key, need to "
                                              "follow the convention 'ID_NUMBER:', probably you are missing a colon."))

                normalized = {}

                for ignored_vuln_id, config in ignore_vulns.items():
                    ignored_vuln_config = config if config else {}

                    if not isinstance(ignored_vuln_config, dict):
                        self.fail(
                            msg.format(hint=f"Wrong configuration under the vulnerability with ID: {ignored_vuln_id}"))

                    context_msg = f'"security" -> "ignore-vulnerabilities" -> "{ignored_vuln_id}" -> '

                    self.fail_if_unrecognized_keys(ignored_vuln_config.keys(), ['reason', 'expires'], param=param,
                                                   ctx=ctx, msg=msg, context_hint=context_msg)

                    reason = ignored_vuln_config.get('reason', '')
                    reason = str(reason) if reason else None
                    expires = ignored_vuln_config.get('expires', '')
                    expires = str(expires) if expires else None

                    try:
                        if int(ignored_vuln_id) < 0:
                            raise ValueError('Negative Vulnerability ID')
                    except ValueError as e:
                        self.fail(msg.format(
                            hint=f"vulnerability id {ignored_vuln_id} under the 'ignore-vulnerabilities' root needs to "
                                 f"be a positive integer")
                        )

                    # Validate expires
                    d = validate_expiration_date(expires)

                    if expires and not d:
                        self.fail(msg.format(hint=f"{context_msg}expires: \"{expires}\" isn't a valid format "
                                                  f"for the expires keyword, "
                                                  "valid options are: YYYY-MM-DD or "
                                                  "YYYY-MM-DD HH:MM:SS")
                                  )

                    normalized[str(ignored_vuln_id)] = {'reason': reason, 'expires': d}

                safety_policy['security']['ignore-vulnerabilities'] = normalized
                safety_policy['filename'] = filename
                safety_policy['raw'] = raw
            else:
                safety_policy['security']['ignore-vulnerabilities'] = {}

            fix_config = safety_policy.get('fix', {})
            self.fail_if_unrecognized_keys(fix_config.keys(), ['auto-security-updates-limit'], param=param, ctx=ctx, msg=msg,
                                           context_hint='"fix" -> ')
            auto_remediation_limit = fix_config.get('auto-security-updates-limit', None)

            if auto_remediation_limit:
                self.fail_if_unrecognized_keys(auto_remediation_limit, ['patch', 'minor', 'major'], param=param, ctx=ctx,
                                               msg=msg,
                                               context_hint='"auto-security-updates-limit" -> ')

            return safety_policy
        except BadParameter as expected_e:
            raise expected_e
        except Exception as e:
            # Don't fail in the default case
            if ctx and isinstance(e, OSError):
                default = ctx.get_parameter_source
                source = default("policy_file") if default("policy_file") else default("policy_file_path")
                if e.errno == 2 and source == click.core.ParameterSource.DEFAULT and value == '.safety-policy.yml':
                    return None

            problem = click.style("Policy file YAML is not valid.")
            hint = click.style("HINT: ", fg='yellow') + str(e)
            self.fail(f"{problem}\n{hint}", param, ctx)

    def shell_complete(
        self, ctx: "Context", param: "Parameter", incomplete: str
    ):
        """Return a special completion marker that tells the completion
        system to use the shell to provide file path completions.

        :param ctx: Invocation context for this command.
        :param param: The parameter that is requesting completion.
        :param incomplete: Value being completed. May be empty.

        .. versionadded:: 8.0
        """
        from click.shell_completion import CompletionItem

        return [CompletionItem(incomplete, type="file")]


class SingletonMeta(type):

    _instances = {}

    _lock = Lock()

    def __call__(cls, *args, **kwargs):
        with cls._lock:
            if cls not in cls._instances:
                instance = super().__call__(*args, **kwargs)
                cls._instances[cls] = instance
        return cls._instances[cls]


class SafetyContext(metaclass=SingletonMeta):
    packages = []
    key = False
    db_mirror = False
    cached = None
    ignore_vulns = None
    ignore_severity_rules = None
    proxy = None
    include_ignored = False
    telemetry = None
    files = None
    stdin = None
    is_env_scan = None
    command: Optional[str] = None
    subcommand: Optional[str] = None
    review = None
    params = {}
    safety_source = 'code'
    local_announcements = []
    scanned_full_path = []
    account = None



def sync_safety_context(f):
    def new_func(*args, **kwargs):
        ctx = SafetyContext()

        legacy_key_added = False
        if "session" in kwargs:
            legacy_key_added = True
            kwargs["key"] = kwargs.get("session").api_key

        for attr in dir(ctx):
            if attr in kwargs:
                setattr(ctx, attr, kwargs.get(attr))

        if legacy_key_added:
            kwargs.pop("key")

        return f(*args, **kwargs)

    return new_func


@sync_safety_context
def get_packages_licenses(*, packages=None, licenses_db=None):
    """Get the licenses for the specified packages based on their version.

    :param packages: packages list
    :param licenses_db: the licenses db in the raw form.
    :return: list of objects with the packages and their respectives licenses.
    """
    SafetyContext().command = 'license'

    if not packages:
        packages = []
    if not licenses_db:
        licenses_db = {}

    packages_licenses_db = licenses_db.get('packages', {})
    filtered_packages_licenses = []

    for pkg in packages:
        # Ignore recursive files not resolved
        if isinstance(pkg, RequirementFile):
            continue
        # normalize the package name
        pkg_name = canonicalize_name(pkg.name)
        # packages may have different licenses depending their version.
        pkg_licenses = packages_licenses_db.get(pkg_name, [])
        if not pkg.version:
            for req in pkg.requirements:
                if is_pinned_requirement(req.specifier):
                    pkg.version = next(iter(req.specifier)).version
                    break
            
            if not pkg.version:
                continue
        version_requested = parse_version(pkg.version)
        license_id = None
        license_name = None
        for pkg_version in pkg_licenses:
            license_start_version = parse_version(pkg_version['start_version'])
            # Stops and return the previous stored license when a new
            # license starts on a version above the requested one.
            if version_requested >= license_start_version:
                license_id = pkg_version['license_id']
            else:
                # We found the license for the version requested
                break

        if license_id:
            license_name = get_license_name_by_id(license_id, licenses_db)
        if not license_id or not license_name:
            license_name = "unknown"

        filtered_packages_licenses.append({
            "package": pkg_name,
            "version": pkg.version,
            "license": license_name
        })

    return filtered_packages_licenses


def get_requirements_content(files):
    requirements_files = {}

    for f in files:
        try:
            f.seek(0)
            requirements_files[f.name] = f.read()
            f.close()
        except Exception as e:
            raise InvalidProvidedReportError(message=f"Unable to read a requirement file scanned in the report. {e}")

    return requirements_files


def is_ignore_unpinned_mode(version):
    ignore = SafetyContext().params.get('ignore_unpinned_requirements')
    return (ignore is None or ignore) and not version


def get_remediations_count(remediations):
    return sum((len(rem.keys()) for pkg, rem in remediations.items()))


def get_hashes(dependency):
    pattern = re.compile(HASH_REGEX_GROUPS)

    return [{'method': method, 'hash': hsh} for method, hsh in
               (pattern.match(d_hash).groups() for d_hash in dependency.hashes)]


def pluralize(word: str, count: int = 0) -> str:
    if count == 1:
        return word
    
    default = {"was": "were", "this": "these", "has": "have"}

    if word in default:
        return default[word]

    if word.endswith("s") or word.endswith("x") or word.endswith("z") \
        or word.endswith("ch") or word.endswith("sh"):
        return word + "es"

    if word.endswith("y"):
        if word[-2] in "aeiou":
            return word + "s"
        else:
            return word[:-1] + "ies"

    return word + "s"


def initializate_config_dirs():
    USER_CONFIG_DIR.mkdir(parents=True, exist_ok=True)

    try:
        SYSTEM_CONFIG_DIR.mkdir(parents=True, exist_ok=True)
    except Exception as e:
        pass


from typing import Optional

from safety.constants import (
    EXIT_CODE_EMAIL_NOT_VERIFIED,
    EXIT_CODE_FAILURE,
    EXIT_CODE_INVALID_API_KEY,
    EXIT_CODE_INVALID_PROVIDED_REPORT,
    EXIT_CODE_INVALID_REQUIREMENT,
    EXIT_CODE_MALFORMED_DB,
    EXIT_CODE_TOO_MANY_REQUESTS,
    EXIT_CODE_UNABLE_TO_FETCH_VULNERABILITY_DB,
    EXIT_CODE_UNABLE_TO_LOAD_LOCAL_VULNERABILITY_DB,
)


class SafetyException(Exception):

    def __init__(self, message="Unhandled exception happened: {info}", info=""):
        self.message = message.format(info=info)
        super().__init__(self.message)

    def get_exit_code(self):
        return EXIT_CODE_FAILURE


class SafetyError(Exception):

    def __init__(self, message="Unhandled Safety generic error", error_code=None):
        self.message = message
        self.error_code = error_code
        super().__init__(self.message)

    def get_exit_code(self):
        return EXIT_CODE_FAILURE


class MalformedDatabase(SafetyError):

    def __init__(self, reason=None, fetched_from="server",
                 message="Sorry, something went wrong.\n" +
                         "Safety CLI can not read the data fetched from {fetched_from} because is malformed.\n"):
        info = "Reason, {reason}".format(reason=reason)
        self.message = message.format(fetched_from=fetched_from) + (info if reason else "")
        super().__init__(self.message)

    def get_exit_code(self):
        return EXIT_CODE_MALFORMED_DB


class DatabaseFetchError(SafetyError):

    def __init__(self, message="Unable to load vulnerability database"):
        self.message = message
        super().__init__(self.message)

    def get_exit_code(self):
        return EXIT_CODE_UNABLE_TO_FETCH_VULNERABILITY_DB


class InvalidProvidedReportError(SafetyError):

    def __init__(self, message="Unable to apply fix: the report needs to be generated from a file. "
                               "Environment isn't supported yet."):
        self.message = message
        super().__init__(self.message)

    def get_exit_code(self):
        return EXIT_CODE_INVALID_PROVIDED_REPORT


class InvalidRequirementError(SafetyError):
    def __init__(self, message="Unable to parse the requirement: {line}", line=""):
        self.message = message.format(line=line)
        super().__init__(self.message)

    def get_exit_code(self):
        return EXIT_CODE_INVALID_REQUIREMENT


class DatabaseFileNotFoundError(DatabaseFetchError):

    def __init__(self, db=None, message="Unable to find vulnerability database in {db}"):
        self.db = db
        self.message = message.format(db=db)
        super().__init__(self.message)

    def get_exit_code(self):
        return EXIT_CODE_UNABLE_TO_LOAD_LOCAL_VULNERABILITY_DB


class InvalidCredentialError(DatabaseFetchError):

    def __init__(self, credential: Optional[str] = None, message="Your authentication credential{credential}is invalid. See {link}.", reason=None):
        self.credential = credential
        self.link = 'https://docs.safetycli.com/safety-docs/support/invalid-api-key-error'
        self.message = message.format(credential=f" '{self.credential}' ", link=self.link) if self.credential else message.format(credential=' ', link=self.link)
        info = f" Reason: {reason}"
        self.message = self.message + (info if reason else "")
        super().__init__(self.message)

    def get_exit_code(self):
        return EXIT_CODE_INVALID_API_KEY

class NotVerifiedEmailError(SafetyError):
    def __init__(self, message="email is not verified"):
        self.message = message
        super().__init__(self.message)

    def get_exit_code(self):
        return EXIT_CODE_EMAIL_NOT_VERIFIED

class TooManyRequestsError(DatabaseFetchError):

    def __init__(self, reason=None,
                 message="Too many requests."):
        info = f" Reason: {reason}"
        self.message = message + (info if reason else "")
        super().__init__(self.message)

    def get_exit_code(self):
        return EXIT_CODE_TOO_MANY_REQUESTS


class NetworkConnectionError(DatabaseFetchError):

    def __init__(self, message="Check your network connection, unable to reach the server."):
        self.message = message
        super().__init__(self.message)


class RequestTimeoutError(DatabaseFetchError):

    def __init__(self, message="Check your network connection, the request timed out."):
        self.message = message
        super().__init__(self.message)


class ServerError(DatabaseFetchError):

    def __init__(self, reason=None,
                 message="Sorry, something went wrong.\n" + "Safety CLI can not connect to the server.\n" +
                         "Our engineers are working quickly to resolve the issue."):
        info = f" Reason: {reason}"
        self.message = message + (info if reason else "")
        super().__init__(self.message)



from enum import Enum
from pathlib import Path
from typing import Optional

from pydantic.dataclasses import dataclass

class FormatMixin:

    @classmethod
    def is_format(cls, format_sub: Optional[Enum], format_instance: Enum):
        """ Check if the value is a variant of the specified format. """
        if not format_sub:
            return False

        if format_sub is format_instance:
            return True

        prefix =  format_sub.value.split('@')[0]
        return prefix == format_instance.value
    
    @property
    def version(self):
        """ Return the version of the format. """
        result = self.value.split('@')

        if len(result) == 2:
            return result[1]
        
        return None


class ScanOutput(FormatMixin, str, Enum):
    JSON = "json"
    SPDX = "spdx"
    SPDX_2_3 = "spdx@2.3"
    SPDX_2_2 = "spdx@2.2"
    HTML = "html"

    SCREEN = "screen"
    NONE = "none"

    def is_silent(self):
        return self in (ScanOutput.JSON, ScanOutput.SPDX, ScanOutput.SPDX_2_3, ScanOutput.SPDX_2_2, ScanOutput.HTML)


class ScanExport(FormatMixin, str, Enum):
    JSON = "json"
    SPDX = "spdx"
    SPDX_2_3 = "spdx@2.3"
    SPDX_2_2 = "spdx@2.2"
    HTML = "html"    

    def get_default_file_name(self, tag: int):
        
        if self is ScanExport.JSON:
            return f"safety-report-{tag}.json"
        elif self in [ScanExport.SPDX, ScanExport.SPDX_2_3, ScanExport.SPDX_2_2]:
            return f"safety-report-spdx-{tag}.json"
        elif self is ScanExport.HTML:
            return f"safety-report-{tag}.html"
        else:
            raise ValueError("Unsupported scan export type")


class SystemScanOutput(str, Enum):
    JSON = "json"
    SCREEN = "screen"

    def is_silent(self):
        return self in (SystemScanOutput.JSON,)   

class SystemScanExport(str, Enum):
    JSON = "json"

@dataclass
class UnverifiedProjectModel():
    id: Optional[str]
    project_path: Path
    created: bool
    name: Optional[str] = None
    url_path: Optional[str] = None    





from pathlib import Path
from safety_schemas.models import Ecosystem, FileType
from typer import FileTextWrite

from .python.main import PythonFile


class InspectableFileContext:
    def __init__(self, file_path: Path, 
                 file_type: FileType) -> None:
        self.file_path = file_path
        self.inspectable_file = None
        self.file_type = file_type

    def __enter__(self): # TODO: Handle permission issue /Applications/...
        try:
            file: FileTextWrite = open(self.file_path, mode='r+') # type: ignore
            self.inspectable_file = TargetFile.create(file_type=self.file_type, file=file)
        except Exception as e:
            # TODO: Report this
            pass
        
        return self.inspectable_file

    def __exit__(self, exc_type, exc_value, traceback):
        if self.inspectable_file:
            self.inspectable_file.file.close()

class TargetFile():

    @classmethod
    def create(cls, file_type: FileType, file: FileTextWrite):
        if file_type.ecosystem == Ecosystem.PYTHON:
            return PythonFile(file=file, file_type=file_type)
        
        raise ValueError("Unsupported ecosystem or file type: " \
                         f"{file_type.ecosystem}:{file_type.value}")



from typing import Optional

from marshmallow import Schema, fields as fields_, post_dump


class CVSSv2(Schema):
    base_score = fields_.Int()
    impact_score = fields_.Int()
    vector_string = fields_.Str()

    class Meta:
        ordered = True


class CVSSv3(Schema):
    base_score = fields_.Int()
    base_severity = fields_.Str()
    impact_score = fields_.Int()
    vector_string = fields_.Str()

    class Meta:
        ordered = True


class VulnerabilitySchemaV05(Schema):
    """
    Legacy JSON report used in Safety 1.10.3
    """

    package_name = fields_.Str()
    vulnerable_spec = fields_.Str()
    version = fields_.Str(attribute='pkg.version')
    advisory = fields_.Str()
    vulnerability_id = fields_.Str()
    cvssv2: Optional[CVSSv2] = fields_.Nested(CVSSv2, attribute='severity.cvssv2')
    cvssv3: Optional[CVSSv3] = fields_.Nested(CVSSv3, attribute='severity.cvssv3')

    class Meta:
        ordered = True

    @post_dump(pass_many=True)
    def wrap_with_envelope(self, data, many, **kwargs):
        return [tuple(d.values()) for d in data]




import json
from collections import namedtuple
from dataclasses import dataclass, field
from datetime import datetime

from dparse.dependencies import Dependency
from dparse import parse, filetypes
from typing import Any, List, Optional
from packaging.specifiers import SpecifierSet
from packaging.requirements import Requirement
from packaging.utils import canonicalize_name
from packaging.version import parse as parse_version

from packaging.version import Version
from safety_schemas.models import ConfigModel

from safety.errors import InvalidRequirementError
from safety_schemas.models import MetadataModel, ReportSchemaVersion, TelemetryModel, \
    PolicyFileModel

try:
    from packaging.version import LegacyVersion as legacyType
except ImportError:
    legacyType = None


class DictConverter(object):

    def to_dict(self, **kwargs):
        pass


announcement_nmt = namedtuple('Announcement', ['type', 'message'])
remediation_nmt = namedtuple('Remediation', ['Package', 'closest_secure_version', 'secure_versions',
                                             'latest_package_version'])
cve_nmt = namedtuple('Cve', ['name', 'cvssv2', 'cvssv3'])
severity_nmt = namedtuple('Severity', ['source', 'cvssv2', 'cvssv3'])
vulnerability_nmt = namedtuple('Vulnerability',
                               ['vulnerability_id', 'package_name', 'pkg', 'ignored', 'ignored_reason', 'ignored_expires',
                                'vulnerable_spec', 'all_vulnerable_specs', 'analyzed_version', 'analyzed_requirement',
                                'advisory', 'is_transitive', 'published_date', 'fixed_versions',
                                'closest_versions_without_known_vulnerabilities', 'resources', 'CVE', 'severity',
                                'affected_versions', 'more_info_url'])
RequirementFile = namedtuple('RequirementFile', ['path'])


class SafetyRequirement(Requirement):
    def __init__(self, requirement: [str, Dependency], found: Optional[str] = None) -> None:
        dep = requirement

        if isinstance(requirement, str):
            deps = parse(requirement, file_type=filetypes.requirements_txt).dependencies
            dep = deps[0] if deps else None

        if not dep:
            raise InvalidRequirementError(line=str(requirement))

        raw_line = dep.line
        to_parse = dep.line
        # Hash and comments are only a pip feature, so removing them.
        if '#' in to_parse:
            to_parse = dep.line.split('#')[0]

        for req_hash in dep.hashes:
            to_parse = to_parse.replace(req_hash, '')

        to_parse = to_parse.replace('\\', '').rstrip()

        try:
            # Try to build a PEP Requirement from the cleaned line
            super(SafetyRequirement, self).__init__(to_parse)
        except Exception:
            raise InvalidRequirementError(line=requirement.line if isinstance(requirement, Dependency) else requirement)

        self.raw = raw_line
        self.found = found

    def __eq__(self, other):
        return str(self) == str(other)
    
    def to_dict(self, **kwargs):
        specifier_obj = self.specifier
        if not "specifier_obj" in kwargs:
            specifier_obj = str(self.specifier)

        return   {
                        'raw': self.raw,
                        'extras': list(self.extras),
                        'marker': str(self.marker) if self.marker else None,
                        'name': self.name,
                        'specifier': specifier_obj,
                        'url': self.url,
                        'found': self.found
                    }


def is_pinned_requirement(spec: SpecifierSet) -> bool:
    if not spec or len(spec) != 1:
        return False

    specifier = next(iter(spec))

    return (specifier.operator == '==' and '*' != specifier.version[-1]) or specifier.operator == '==='


@dataclass
class Package(DictConverter):
    name: str
    version: Optional[str]
    requirements: List[SafetyRequirement]
    found: Optional[str] = None
    absolute_path: Optional[str] = None
    insecure_versions: List[str] = field(default_factory=lambda: [])
    secure_versions: List[str] = field(default_factory=lambda: [])
    latest_version_without_known_vulnerabilities: Optional[str] = None
    latest_version: Optional[str] = None
    more_info_url: Optional[str] = None

    def has_unpinned_req(self):
        for req in self.requirements:
            if not is_pinned_requirement(req.specifier):
                return True
        return False

    def get_unpinned_req(self):
        return filter(lambda r: not is_pinned_requirement(r.specifier), self.requirements)

    def filter_by_supported_versions(self, versions: [str]) -> [str]:
        allowed = []

        for version in versions:
            try:
                parse_version(version)
                allowed.append(version)
            except Exception:
                pass

        return allowed

    def get_versions(self, db_full):
        pkg_meta = db_full.get('meta', {}).get('packages', {}).get(self.name, {})
        versions = self.filter_by_supported_versions(
            pkg_meta.get("insecure_versions", []) + pkg_meta.get("secure_versions", []))
        return set(versions)

    def refresh_from(self, db_full):
        base_domain = db_full.get('meta', {}).get('base_domain')
        pkg_meta = db_full.get('meta', {}).get('packages', {}).get(canonicalize_name(self.name), {})

        kwargs = {'insecure_versions': self.filter_by_supported_versions(pkg_meta.get("insecure_versions", [])),
                  'secure_versions': self.filter_by_supported_versions(pkg_meta.get("secure_versions", [])),
                  'latest_version_without_known_vulnerabilities': pkg_meta.get("latest_secure_version",
                                                                               None),
                  'latest_version': pkg_meta.get("latest_version", None),
                  'more_info_url': f"{base_domain}{pkg_meta.get('more_info_path', '')}"}

        self.update(kwargs)

    def to_dict(self, **kwargs):
        if kwargs.get('short_version', False):
            return {
                'name': self.name,
                'version': self.version,
                'requirements': self.requirements
            }

        return {'name': self.name,
                'version': self.version,
                'requirements': self.requirements,
                'found': None,
                'insecure_versions': self.insecure_versions,
                'secure_versions': self.secure_versions,
                'latest_version_without_known_vulnerabilities': self.latest_version_without_known_vulnerabilities,
                'latest_version': self.latest_version,
                'more_info_url': self.more_info_url
                }

    def update(self, new):
        for key, value in new.items():
            if hasattr(self, key):
                setattr(self, key, value)


class Announcement(announcement_nmt):
    pass


class Remediation(remediation_nmt, DictConverter):

    def to_dict(self):
        return {'package': self.Package.name,
                'closest_secure_version': self.closest_secure_version,
                'secure_versions': self.secure_versions,
                'latest_package_version': self.latest_package_version
                }


@dataclass
class Fix:
    dependency: Any = None
    previous_version: Any = None
    previous_spec: Optional[str] = None
    other_options: [str] = field(default_factory=lambda: [])
    updated_version: Any = None
    update_type: str = ""
    package: str = ""
    status: str = ""
    applied_at: str = ""
    fix_type: str = ""
    more_info_url: str = ""


class CVE(cve_nmt, DictConverter):

    def to_dict(self):
        return {'name': self.name, 'cvssv2': self.cvssv2, 'cvssv3': self.cvssv3}


class Severity(severity_nmt, DictConverter):
    def to_dict(self):
        result = {'severity': {'source': self.source}}

        result['severity']['cvssv2'] = self.cvssv2
        result['severity']['cvssv3'] = self.cvssv3

        return result


class SafetyEncoder(json.JSONEncoder):
    def default(self, value):
        if isinstance(value, SafetyRequirement):
            return value.to_dict()
        elif isinstance(value, Version) or (legacyType and isinstance(value, legacyType)):
            return str(value)
        else:
            return super().default(value)


class Vulnerability(vulnerability_nmt):

    def to_dict(self):
        empty_list_if_none = ['fixed_versions', 'closest_versions_without_known_vulnerabilities', 'resources']
        result = {
        }

        ignore = ['pkg']

        for field, value in zip(self._fields, self):
            if field in ignore:
                continue

            if value is None and field in empty_list_if_none:
                value = []

            if isinstance(value, set):
                result[field] = list(value)
            elif isinstance(value, CVE):
                val = None
                if value.name.startswith("CVE"):
                    val = value.name
                result[field] = val
            elif isinstance(value, DictConverter):
                result.update(value.to_dict())
            elif isinstance(value, SpecifierSet) or isinstance(value, datetime):
                result[field] = str(value)
            else:
                result[field] = value

        return result

    def get_advisory(self):
        return self.advisory.replace('\r', '') if self.advisory else "No advisory found for this vulnerability."
    
    def to_model_dict(self):
        try:
            affected_spec = next(iter(self.vulnerable_spec))
        except Exception:
            affected_spec = ""

        repr = {
            "id": self.vulnerability_id,
            "package_name": self.package_name,
            "vulnerable_spec": affected_spec,
            "analyzed_specification": self.analyzed_requirement.raw
        }

        if self.ignored:
            repr["ignored"] = {"reason": self.ignored_reason, 
                               "expires": self.ignored_expires}

        return repr


@dataclass
class Safety:
    client: Any
    keys: Any


from safety.auth.models import Auth
from rich.console import Console

@dataclass
class SafetyCLI:
    auth: Optional[Auth] = None
    telemetry: Optional[TelemetryModel] = None
    metadata: Optional[MetadataModel] = None
    schema: Optional[ReportSchemaVersion] = None
    project = None
    config: Optional[ConfigModel] = None
    console: Optional[Console] = None
    system_scan_policy: Optional[PolicyFileModel] = None
    platform_enabled: bool = False



import unittest
from unittest import mock
from unittest.mock import MagicMock, Mock, patch
from pathlib import Path
import datetime

from safety.scan.render import print_announcements, print_ignore_details, render_header
from safety_schemas.models import ProjectModel, IgnoreCodes

class TestRender(unittest.TestCase):
    @patch('safety.scan.render.get_safety_version')
    def test_render_header(self, mock_get_safety_version):
        mock_get_safety_version.return_value = '3.0.0'

        datetime_mock = Mock(wraps=datetime.datetime)
        datetime_mock.now.return_value = datetime.datetime(2025, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc)

        with patch('datetime.datetime', new=datetime_mock) as mock_now:
            mock_now.return_value = datetime.datetime(2025, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc)

            # Project scan
            targets = [Path('/target1'), Path('/target2')]
            expected_result = f'Safety 3.0.0 scanning {targets[0]}, {targets[1]}\n2025-01-01 00:00:00 UTC'
            self.assertEqual(str(render_header(targets, False)), expected_result)

            # System Scan
            expected_result = 'Safety 3.0.0 running system scan\n2025-01-01 00:00:00 UTC'
            self.assertEqual(str(render_header(targets, True)), expected_result)

    @patch('safety.scan.render.safety.get_announcements')
    @patch('safety.scan.render.get_basic_announcements')
    @patch('safety.scan.render.Console')
    def test_print_announcements(self, mock_console, mock_get_basic_announcements, mock_get_announcements):

        mock_get_announcements.return_value = [
            {'type': 'info', 'message': 'Info message'},
            {'type': 'warning', 'message': 'Warning message'},
            {'type': 'error', 'message': 'Error message'},
        ]

        mock_get_basic_announcements.return_value = mock_get_announcements.return_value
        console = mock_console.return_value

        ctx = MagicMock()
        ctx.obj.auth.client = MagicMock()
        ctx.obj.config.telemetry_enabled = False
        ctx.obj.telemetry = MagicMock()

        print_announcements(console, ctx)

        console.print.assert_any_call()
        console.print.assert_any_call("[bold]Safety Announcements:[/bold]")
        console.print.assert_any_call()
        console.print.assert_any_call("[default]* Info message[/default]")
        console.print.assert_any_call("[yellow]* Warning message[/yellow]")
        console.print.assert_any_call("[red]* Error message[/red]")    


    @patch('safety.scan.render.render_to_console')
    def test_print_ignore_details(self, render_to_console_mocked):
        render_to_console_mocked.return_value = "render_to_console_mocked"
        from safety.console import main_console
        console = MagicMock(wraps=main_console)
        console.print = MagicMock()

        # Create a fake project
        project = ProjectModel(id='prj-id')

        # Create a fake ignored vulnerabilities data
        ignored_vulns_data = [
            MagicMock(ignored_code=IgnoreCodes.manual.value, vulnerability_id='v1', package_name='p1'),
            MagicMock(ignored_code=IgnoreCodes.cvss_severity.value, vulnerability_id='v2', package_name='p2'),
            MagicMock(ignored_code=IgnoreCodes.unpinned_specification.value, vulnerability_id='v3', package_name='p3'),
            MagicMock(ignored_code=IgnoreCodes.environment_dependency.value, vulnerability_id='v4', package_name='p4'),
        ]

        # Call the function
        print_ignore_details(console, project, [], True, ignored_vulns_data)

        # Check that the console.print method was called with the expected arguments
        console.print.assert_any_call("[number]1[/number] were manually ignored due to the project policy:")
        console.print.assert_any_call("[number]1[/number] vulnerability was ignored because of their severity or exploitability impacted the following package: p2")
        console.print.assert_any_call("[number]1[/number] vulnerability was ignored because they are inside an environment dependency.")
        console.print.assert_any_call("[number]1[/number] vulnerability was ignored because this package has unpinned specs: p3")




import os
from pathlib import Path
from typing import Optional, Tuple
import typer
from safety.scan.main import save_project_info
from safety.scan.models import ScanExport, ScanOutput, UnverifiedProjectModel
from safety.scan.render import print_wait_project_verification, prompt_project_id, prompt_link_project

from safety_schemas.models import AuthenticationType, ProjectModel, Stage


MISSING_SPDX_EXTENSION_MSG = "spdx extra is not installed, please install it with: pip install safety[spdx]"


def raise_if_not_spdx_extension_installed():
    try:
        import spdx_tools.spdx
    except Exception as e:
        raise typer.BadParameter(MISSING_SPDX_EXTENSION_MSG)    


def save_as_callback(save_as: Optional[Tuple[ScanExport, Path]]):
    export_type, export_path = save_as if save_as else (None, None)

    if ScanExport.is_format(export_type, ScanExport.SPDX):
        raise_if_not_spdx_extension_installed()

    return (export_type.value, export_path) if export_type and export_path else (export_type, export_path)

def output_callback(output: ScanOutput):

    if ScanOutput.is_format(output, ScanExport.SPDX):
        raise_if_not_spdx_extension_installed()
    
    return output.value


def fail_if_not_allowed_stage(ctx: typer.Context):
    if ctx.resilient_parsing:
        return
    
    stage = ctx.obj.auth.stage
    auth_type: AuthenticationType = ctx.obj.auth.client.get_authentication_type()

    if os.getenv("SAFETY_DB_DIR"):
        return

    if not auth_type.is_allowed_in(stage):
        raise typer.BadParameter(f"'{auth_type.value}' auth type isn't allowed with " \
                                 f"the '{stage}' stage.")


def save_verified_project(ctx, slug, name, project_path, url_path):
    ctx.obj.project = ProjectModel(
        id=slug,
        name=name,
        project_path=project_path,
        url_path=url_path
    )
    if ctx.obj.auth.stage is Stage.development:
        save_project_info(project=ctx.obj.project, 
                          project_path=project_path)


def check_project(console, ctx, session,
                   unverified_project: UnverifiedProjectModel, 
                   stage, 
                   git_origin, ask_project_id=False):
    stage = ctx.obj.auth.stage
    source = ctx.obj.telemetry.safety_source if ctx.obj.telemetry else None
    data = {"scan_stage": stage, "safety_source": source}

    PRJ_SLUG_KEY = "project_slug"
    PRJ_SLUG_SOURCE_KEY = "project_slug_source"
    PRJ_GIT_ORIGIN_KEY = "git_origin"

    if git_origin:
        data[PRJ_GIT_ORIGIN_KEY] = git_origin

    if unverified_project.id:
        data[PRJ_SLUG_KEY] = unverified_project.id
        data[PRJ_SLUG_SOURCE_KEY] = ".safety-project.ini"
    elif not git_origin or ask_project_id:
        # Set a project id for this scan (no spaces). If empty Safety will use: pyupio:
        parent_root_name = None
        if unverified_project.project_path.parent.name:
            parent_root_name = unverified_project.project_path.parent.name

        unverified_project.id = prompt_project_id(console, stage, parent_root_name)
        data[PRJ_SLUG_KEY] = unverified_project.id
        data[PRJ_SLUG_SOURCE_KEY] = "user"

    status = print_wait_project_verification(console, data[PRJ_SLUG_KEY] if data.get(PRJ_SLUG_KEY, None) else "-", 
                                    (session.check_project, data), on_error_delay=1)

    return status


def verify_project(console, ctx, session,
                   unverified_project: UnverifiedProjectModel, 
                   stage, 
                   git_origin):
    
    verified_prj = False

    link_prj = True

    while not verified_prj:
        result = check_project(console, ctx, session, unverified_project, stage, git_origin, ask_project_id=not link_prj)

        unverified_slug = result.get("slug")

        project = result.get("project", None)
        user_confirm = result.get("user_confirm", False)

        if user_confirm:
            if project and link_prj:
                prj_name = project.get("name", None)
                prj_admin_email = project.get("admin", None)

                link_prj = prompt_link_project(prj_name=prj_name,
                                    prj_admin_email=prj_admin_email,
                                    console=console)
                
                if not link_prj:
                    continue

        verified_prj = print_wait_project_verification(
            console, unverified_slug, (session.project, 
                                       {"project_id": unverified_slug}),
                                       on_error_delay=1)
        
        if verified_prj and isinstance(verified_prj, dict) and verified_prj.get("slug", None):
            save_verified_project(ctx, verified_prj["slug"], verified_prj.get("name", None), 
                                  unverified_project.project_path, verified_prj.get("url", None))
        else:
            verified_prj = False
